# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the core module.
"""

from contextlib import nullcontext
from unittest.mock import patch

import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.models import Gaussian2D
from astropy.utils.exceptions import (AstropyDeprecationWarning,
                                      AstropyUserWarning)
from numpy.testing import assert_allclose, assert_array_equal

from photutils.centroids.core import (CentroidQuadratic, centroid_com,
                                      centroid_quadratic, centroid_sources)
from photutils.centroids.gaussian import centroid_1dg, centroid_2dg
from photutils.datasets import make_4gaussians_image, make_noise_image


def _make_gaussian_source(shape, amplitude, xc, yc, xstd, ystd, theta):
    """
    Make a 2D Gaussian source.
    """
    yy, xx = np.mgrid[0:shape[0], 0:shape[1]]
    model = Gaussian2D(amplitude, xc, yc, xstd, ystd, theta)
    return model(xx, yy)


@pytest.fixture(name='test_data')
def fixture_test_data():
    """
    Create test data with multiple Gaussian sources.
    """
    ysize = 50
    xsize = 47
    yy, xx = np.mgrid[0:ysize, 0:xsize]
    data = np.zeros((ysize, xsize))
    xpos = (1, 25, 25, 35, 46)
    ypos = (1, 25, 12, 35, 49)
    for xc, yc in zip(xpos, ypos, strict=True):
        model = Gaussian2D(10.0, xc, yc, x_stddev=2, y_stddev=2,
                           theta=0)
        data += model(xx, yy)
    return data, xpos, ypos


@pytest.fixture(name='nan_data')
def fixture_nan_data():
    """
    Create Gaussian test data with NaN values in row 20.
    """
    xc_ref = 24.7
    yc_ref = 25.2
    data = _make_gaussian_source((50, 50), 2.4, xc_ref, yc_ref, 5.0, 5.0, 0)
    data[20, :] = np.nan
    return data, xc_ref, yc_ref


@pytest.mark.parametrize('x_std', [3.2, 4.0])
@pytest.mark.parametrize('y_std', [5.7, 4.1])
@pytest.mark.parametrize('theta', np.deg2rad([30.0, 45.0]))
@pytest.mark.parametrize('units', [True, False])
def test_centroid_com(x_std, y_std, theta, units):
    """
    Test centroid_com with Gaussian data.
    """
    xc_ref = 25.7
    yc_ref = 26.2
    data = _make_gaussian_source((50, 47), 2.4, xc_ref, yc_ref, x_std, y_std,
                                 theta)

    if units:
        data = data * u.nJy

    xc, yc = centroid_com(data)
    assert_allclose((xc, yc), (xc_ref, yc_ref), rtol=0, atol=1.0e-3)

    # Test with mask
    x0 = 11
    y0 = 15
    data[y0, x0] = 1.0e5 * u.nJy if units else 1.0e5
    mask = np.zeros(data.shape, dtype=bool)
    mask[y0, x0] = True
    xc, yc = centroid_com(data, mask=mask)
    assert_allclose((xc, yc), (xc_ref, yc_ref), rtol=0, atol=1.0e-3)


@pytest.mark.parametrize('use_mask', [True, False])
def test_centroid_com_nan_withmask(nan_data, use_mask):
    """
    Test centroid_com with NaN values and optional mask.
    """
    data, xc_ref, yc_ref = nan_data
    if use_mask:
        mask = np.zeros(data.shape, dtype=bool)
        mask[20, :] = True
        ctx = nullcontext()
    else:
        mask = None
        match = 'Input data contains non-finite values'
        ctx = pytest.warns(AstropyUserWarning, match=match)

    with ctx as warnlist:
        xc, yc = centroid_com(data, mask=mask)
        assert_allclose(xc, xc_ref, rtol=0, atol=1.0e-3)
        assert yc > yc_ref
        if not use_mask:
            assert len(warnlist) == 1


def test_centroid_com_allmask():
    """
    Test centroid_com when all data are masked or zero.
    """
    xc_ref = 24.7
    yc_ref = 25.2
    data = _make_gaussian_source((50, 50), 2.4, xc_ref, yc_ref, 5.0, 5.0, 0)

    mask = np.ones(data.shape, dtype=bool)
    xc, yc = centroid_com(data, mask=mask)
    assert np.isnan(xc)
    assert np.isnan(yc)

    data = np.zeros((25, 25))
    xc, yc = centroid_com(data, mask=None)
    assert np.isnan(xc)
    assert np.isnan(yc)


def test_centroid_com_invalid_inputs():
    """
    Test centroid_com with invalid inputs.
    """
    data = np.zeros((4, 4))
    mask = np.zeros((2, 2), dtype=bool)
    match = 'data and mask must have the same shape'
    with pytest.raises(ValueError, match=match):
        centroid_com(data, mask=mask)


@pytest.mark.parametrize('ndim', [1, 2, 3, 4, 5])
def test_centroid_com_zero_sum(ndim):
    """
    Test centroid_com when the sum of the data is zero, which should
    return NaN.
    """
    data = np.zeros([10] * ndim)
    cen = centroid_com(data)
    assert cen.shape == (ndim,)
    for cen_ in cen:
        assert np.isnan(cen_)


def test_centroid_com_masked_array():
    """
    Test centroid_com with a MaskedArray input.
    """
    data = np.ma.array([[1.0, 1.0, 1.0],
                        [1.0, 100.0, 1.0],
                        [10.0, 1.0, 1.0]],
                       mask=[[0, 0, 0],
                             [0, 1, 0],
                             [0, 0, 0]])

    # If mask is respected, peak (1, 1) is ignored and centroid will be
    # pulled towards (0, 0).
    xc1, yc1 = centroid_com(data)

    # Compare with explicit mask
    xc2, yc2 = centroid_com(data.data, mask=data.mask)
    assert xc1 == xc2
    assert yc1 == yc2

    # Combined mask test
    mask_arg = np.zeros(data.shape, dtype=bool)
    mask_arg[0, 0] = True
    # Now both (1, 1) and (0, 0) are masked.
    xc3, yc3 = centroid_com(data, mask=mask_arg)

    full_mask = data.mask | mask_arg
    xc4, yc4 = centroid_com(data.data, mask=full_mask)
    assert xc3 == xc4
    assert yc3 == yc4


def test_centroid_com_mutation():
    """
    Test that centroid_com does not mutate the input data or mask.
    """
    data = np.ones((5, 5))
    mask = np.zeros((5, 5), dtype=bool)
    mask[2, 2] = True
    data_orig = data.copy()
    mask_orig = mask.copy()
    centroid_com(data, mask=mask)
    assert_array_equal(data, data_orig)
    assert_array_equal(mask, mask_orig)


@pytest.mark.parametrize('x_std', [3.2, 4.0])
@pytest.mark.parametrize('y_std', [5.7, 4.1])
@pytest.mark.parametrize('theta', np.deg2rad([30.0, 45.0]))
@pytest.mark.parametrize('units', [True, False])
def test_centroid_quadratic(x_std, y_std, theta, units):
    """
    Test centroid_quadratic with Gaussian data.
    """
    xc_ref = 25.7
    yc_ref = 26.2
    data = _make_gaussian_source((50, 47), 2.4, xc_ref, yc_ref, x_std, y_std,
                                 theta)

    if units:
        data = data * u.nJy

    xc, yc = centroid_quadratic(data)
    assert_allclose((xc, yc), (xc_ref, yc_ref), rtol=0, atol=0.015)

    # Test with mask
    x0 = 11
    y0 = 15
    data[y0, x0] = 1.0e5 * u.nJy if units else 1.0e5
    mask = np.zeros(data.shape, dtype=bool)
    mask[y0, x0] = True
    data[y0, x0] = 1.0e5 * u.nJy if units else 1.0e5
    xc, yc = centroid_quadratic(data, mask=mask)
    assert_allclose((xc, yc), (xc_ref, yc_ref), rtol=0, atol=0.015)


def test_centroid_quadratic_xypeak():
    """
    Test centroid_quadratic with xpeak and ypeak inputs.
    """
    data = np.zeros((11, 11))
    data[5, 5] = 100
    data[7, 7] = 110
    data[9, 9] = 120

    xycen1 = centroid_quadratic(data, fit_boxsize=3)
    assert_allclose(xycen1, (9, 9))

    with pytest.warns(AstropyDeprecationWarning):
        xycen2 = centroid_quadratic(data, xpeak=5, ypeak=5, fit_boxsize=3)
    assert_allclose(xycen2, (5, 5))

    with pytest.warns(AstropyDeprecationWarning):
        xycen3 = centroid_quadratic(data, xpeak=5, ypeak=5, fit_boxsize=3,
                                    search_boxsize=5)
    assert_allclose(xycen3, (7, 7))

    match = 'xpeak is outside the input data'
    with (pytest.warns(AstropyDeprecationWarning),
          pytest.raises(ValueError, match=match)):
        centroid_quadratic(data, xpeak=15, ypeak=5)
    with (pytest.warns(AstropyDeprecationWarning),
          pytest.raises(ValueError, match=match)):
        centroid_quadratic(data, xpeak=15, ypeak=15)

    match = 'ypeak is outside the input data'
    with (pytest.warns(AstropyDeprecationWarning),
          pytest.raises(ValueError, match=match)):
        centroid_quadratic(data, xpeak=5, ypeak=15)


def test_centroid_quadratic_nan():
    """
    Test centroid_quadratic with NaN values.
    """
    data = _make_gaussian_source((100, 100), 42.1, 47.8, 52.4, 4.7, 4.7, 0)
    error = make_noise_image(data.shape, mean=0., stddev=2.4, seed=123)
    data += error

    data[50, 50] = np.nan
    mask = ~np.isfinite(data)
    xycen = centroid_quadratic(data, mask=mask)
    assert_allclose(xycen, [47.58324, 51.827182])


@pytest.mark.parametrize('use_mask', [True, False])
def test_centroid_quadratic_nan_withmask(nan_data, use_mask):
    """
    Test centroid_quadratic with NaN values and optional mask.
    """
    data, xc_ref, yc_ref = nan_data
    if use_mask:
        mask = np.zeros(data.shape, dtype=bool)
        mask[20, :] = True
        ctx = nullcontext()
    else:
        mask = None
        match = 'Input data contains non-finite values'
        ctx = pytest.warns(AstropyUserWarning, match=match)

    with ctx as warnlist:
        xc, yc = centroid_quadratic(data, mask=mask)
        assert_allclose(xc, xc_ref, rtol=0, atol=0.15)
        assert_allclose(yc, yc_ref, rtol=0, atol=0.15)
        if not use_mask:
            assert len(warnlist) == 1


def test_centroid_quadratic_nan_in_fitbox():
    """
    Test centroid_quadratic with a NaN inside the fit box.

    This tests that non-finite values are removed from the coefficient
    matrix and cutout before the least-squares fit. The NaN pixel is
    masked so that _process_data_mask fills it with NaN (fill_value),
    which is then filtered out by the ``finite_mask`` check inside the
    fit.
    """
    data = _make_gaussian_source((11, 11), 100.0, 5.0, 5.0, 2.0, 2.0, 0)
    # Place a NaN adjacent to the peak; with fit_boxsize=5 centered at
    # (5, 5) the fit box covers rows/cols [3:8], so (row=5, col=4) is
    # inside the box and will trigger the ``if not np.all(finite_mask)``
    # branch.
    data[5, 4] = np.nan
    mask = np.zeros(data.shape, dtype=bool)
    mask[5, 4] = True  # suppress the non-finite warning via explicit mask
    xycen = centroid_quadratic(data, mask=mask, fit_boxsize=5)
    assert_allclose(xycen, (5.0, 5.0), atol=0.01)


def test_centroid_quadratic_npts():
    """
    Test centroid_quadratic with insufficient unmasked data points.
    """
    data = np.zeros((3, 3))
    data[1, 1] = 1
    mask = np.zeros(data.shape, dtype=bool)
    mask[0, :] = True
    mask[2, :] = True
    match = 'at least 6 unmasked data points'
    with pytest.warns(AstropyUserWarning, match=match):
        centroid_quadratic(data, mask=mask)


def test_centroid_quadratic_invalid_inputs():
    """
    Test centroid_quadratic with invalid inputs.
    """
    data = np.zeros((4, 4, 4))
    match = 'data must be a 2D array'
    with pytest.raises(ValueError, match=match):
        centroid_quadratic(data)

    data = np.zeros((4, 4))
    mask = np.zeros((2, 2), dtype=bool)
    match = 'xpeak and ypeak must both be input or "None"'
    with (pytest.warns(AstropyDeprecationWarning),
          pytest.raises(ValueError, match=match)):
        centroid_quadratic(data, xpeak=3, ypeak=None)
    with (pytest.warns(AstropyDeprecationWarning),
          pytest.raises(ValueError, match=match)):
        centroid_quadratic(data, xpeak=None, ypeak=3)

    match = 'fit_boxsize must have 1 or 2 elements'
    with pytest.raises(ValueError, match=match):
        centroid_quadratic(data, fit_boxsize=(2, 2, 2))

    match = 'fit_boxsize must have an odd value for both axes'
    with pytest.raises(ValueError, match=match):
        centroid_quadratic(data, fit_boxsize=(-2, 2))
    with pytest.raises(ValueError, match=match):
        centroid_quadratic(data, fit_boxsize=(2, 2))

    match = 'data and mask must have the same shape'
    with pytest.raises(ValueError, match=match):
        centroid_quadratic(data, mask=mask)


def test_centroid_quadratic_edge():
    """
    Test centroid_quadratic when the maximum is at the edge.
    """
    data = np.zeros((11, 11))
    data[1, 1] = 100
    data[9, 9] = 100

    with pytest.warns(AstropyDeprecationWarning):
        xycen = centroid_quadratic(data, xpeak=1, ypeak=1, fit_boxsize=5)
    assert_allclose(xycen, (0.923077, 0.923077))

    with pytest.warns(AstropyDeprecationWarning):
        xycen = centroid_quadratic(data, xpeak=9, ypeak=9, fit_boxsize=5)
    assert_allclose(xycen, (9.076923, 9.076923))

    data = np.zeros((5, 5))
    data[0, 0] = 100
    match = 'maximum value is at the edge'
    with pytest.warns(AstropyUserWarning, match=match):
        xycen = centroid_quadratic(data)
    assert_allclose(xycen, (0, 0))


def test_centroid_quadratic_mutation():
    """
    Test that centroid_quadratic does not mutate the input data or mask.
    """
    data = np.ones((11, 11))
    data[5, 5] = 10.0
    mask = np.zeros((11, 11), dtype=bool)
    mask[0, 0] = True
    data_orig = data.copy()
    mask_orig = mask.copy()
    centroid_quadratic(data, mask=mask)
    assert_array_equal(data, data_orig)
    assert_array_equal(mask, mask_orig)


def test_centroid_quadratic_units():
    """
    Test that centroid_quadratic strips Quantity units and returns the
    same result as a plain float array.
    """
    xc_ref = 25.7
    yc_ref = 26.2
    data = _make_gaussian_source((50, 47), 2.4, xc_ref, yc_ref, 3.2, 5.7, 0)

    xc_plain, yc_plain = centroid_quadratic(data)
    xc_unit, yc_unit = centroid_quadratic(data * u.nJy)
    assert_allclose(xc_plain, xc_unit)
    assert_allclose(yc_plain, yc_unit)


def test_centroid_quadratic_fit_failed():
    """
    Test centroid_quadratic when the quadratic fit fails.

    This tests the LinAlgError exception handling. Since lstsq is
    very robust (uses SVD internally), we use mocking to trigger this
    condition.
    """
    data = np.zeros((11, 11))
    data[5, 5] = 10.0
    data[4, 5] = 8.0
    data[6, 5] = 8.0
    data[5, 4] = 8.0
    data[5, 6] = 8.0

    with patch('numpy.linalg.lstsq', side_effect=np.linalg.LinAlgError):
        match = 'quadratic fit failed'
        with (pytest.warns(AstropyDeprecationWarning),
              pytest.warns(AstropyUserWarning, match=match)):
            xycen = centroid_quadratic(data, xpeak=5, ypeak=5, fit_boxsize=5)
        assert np.isnan(xycen[0])
        assert np.isnan(xycen[1])


def test_centroid_quadratic_no_maximum():
    """
    Test centroid_quadratic when the quadratic fit does not have a
    maximum.

    This tests the case where the fitted polynomial has a saddle point
    or minimum instead of a maximum (det <= 0 or positive curvature).
    """
    # Create data with a saddle-like pattern that will result in a
    # quadratic fit without a proper maximum
    data = np.zeros((11, 11))
    y, x = np.mgrid[0:11, 0:11]

    # Create a saddle: z = x^2 - y^2 (positive curvature in x, negative
    # in y)
    data = (x - 5.0)**2 - (y - 5.0)**2 + 10
    # Add a peak so the fit box centers there
    data[5, 5] = 20.0

    match = 'quadratic fit does not have a maximum'
    with (pytest.warns(AstropyDeprecationWarning),
          pytest.warns(AstropyUserWarning, match=match)):
        xycen = centroid_quadratic(data, xpeak=5, ypeak=5, fit_boxsize=5)
    assert np.isnan(xycen[0])
    assert np.isnan(xycen[1])


def test_centroid_quadratic_max_outside_image():
    """
    Test centroid_quadratic when the polynomial maximum falls outside
    the image.

    This tests the case where the quadratic fit has a valid maximum but
    it lies outside the image boundaries.
    """
    # Create data where values increase toward the origin (0, 0) but
    # with a local peak at (3, 3). This causes the quadratic fit to
    # extrapolate the maximum outside the image boundaries.
    data = np.zeros((7, 7))
    y, x = np.mgrid[0:7, 0:7]
    data = 10 - x.astype(float) - y.astype(float)
    data = np.maximum(data, 0.1)
    data[3, 3] = 6.0  # local peak to center the fit

    match = 'quadratic polynomial maximum value falls outside'
    with (pytest.warns(AstropyDeprecationWarning),
          pytest.warns(AstropyUserWarning, match=match)):
        xycen = centroid_quadratic(data, xpeak=3, ypeak=3, fit_boxsize=5)
    assert np.isnan(xycen[0])
    assert np.isnan(xycen[1])


class TestCentroidSources:
    """
    Test the centroid_sources function.
    """

    @staticmethod
    def test_centroid_sources():
        """
        Test centroid_sources with Gaussian data.
        """
        theta = np.pi / 6.0
        data = _make_gaussian_source((50, 47), 2.4, 25.7, 26.2, 3.2, 5.7,
                                     theta)
        error = np.ones(data.shape, dtype=float)
        mask = np.zeros(data.shape, dtype=bool)
        mask[10, 10] = True
        xpos = [25.0]
        ypos = [26.0]
        xc, yc = centroid_sources(data, xpos, ypos, box_size=21, mask=mask)
        assert_allclose(xc, (25.67,), atol=1e-1)
        assert_allclose(yc, (26.18,), atol=1e-1)

        xc, yc = centroid_sources(data, xpos, ypos, error=error, box_size=11,
                                  centroid_func=centroid_1dg)
        assert_allclose(xc, (25.67,), atol=1e-1)
        assert_allclose(yc, (26.41,), atol=1e-1)

        match = 'xpos must be a 1D array'
        with pytest.raises(ValueError, match=match):
            centroid_sources(data, [[25]], 26, box_size=11)
        match = 'ypos must be a 1D array'
        with pytest.raises(ValueError, match=match):
            centroid_sources(data, 25, [[26]], box_size=11)
        match = 'xpos and ypos must have the same length'
        with pytest.raises(ValueError, match=match):
            centroid_sources(data, [25, 26], [26], box_size=11)
        match = 'box_size must have 1 or 2 elements'
        with pytest.raises(ValueError, match=match):
            centroid_sources(data, 25, 26, box_size=(1, 2, 3))
        match = 'box_size or footprint must be defined'
        with pytest.raises(ValueError, match=match):
            centroid_sources(data, 25, 26, box_size=None, footprint=None)
        match = 'footprint must be a 2D array'
        with pytest.raises(ValueError, match=match):
            centroid_sources(data, 25, 26, footprint=np.ones((3, 3, 3)))

        def test_func(data):
            return data

        match = "The input 'centroid_func' must have a 'mask' keyword"
        with pytest.raises(ValueError, match=match):
            centroid_sources(data, [25], 26, centroid_func=test_func)

        match = 'data must be a 2D array'
        with pytest.raises(ValueError, match=match):
            centroid_sources(np.ones((3, 3, 3)), 1, 1, box_size=3)

    @pytest.mark.parametrize('centroid_func', [centroid_com,
                                               centroid_quadratic,
                                               centroid_1dg, centroid_2dg])
    def test_xypos(self, test_data, centroid_func):
        """
        Test centroid_sources with xpos/ypos outside data range.
        """
        data = test_data[0]
        match = 'xpos, ypos values contain points outside the input data'
        with pytest.raises(ValueError, match=match):
            centroid_sources(data, 47, 50, box_size=5,
                             centroid_func=centroid_func)

    def test_gaussian_fits_npts(self, test_data):
        """
        Test centroid_sources with Gaussian fits with insufficient
        points.
        """
        data, xpos, ypos = test_data
        xcen, ycen = centroid_sources(data, xpos, ypos, box_size=3,
                                      centroid_func=centroid_1dg)

        xres = np.copy(xpos).astype(float)
        yres = np.copy(ypos).astype(float)
        xres[-1] = 46.689208
        yres[-1] = 49.689208
        assert_allclose(xcen, xres, atol=1e-5)
        assert_allclose(ycen, yres, atol=1e-5)

        match = 'Centroid failed for source'
        with pytest.warns(AstropyUserWarning, match=match):
            xcen, ycen = centroid_sources(data, xpos, ypos, box_size=3,
                                          centroid_func=centroid_2dg)
        xres[-1] = np.nan
        yres[-1] = np.nan
        assert_allclose(xcen, xres)
        assert_allclose(ycen, yres)

        xcen, ycen = centroid_sources(data, xpos, ypos, box_size=5,
                                      centroid_func=centroid_1dg)
        assert_allclose(xcen, xpos)
        assert_allclose(ycen, ypos)

        match = 'Centroid failed for source'
        with pytest.warns(AstropyUserWarning, match=match):
            xcen, ycen = centroid_sources(data, xpos, ypos, box_size=3,
                                          centroid_func=centroid_quadratic)
        assert_allclose(xcen, xres)
        assert_allclose(ycen, yres)

    def test_centroid_quadratic_mask(self):
        """
        Test centroid_sources with centroid_quadratic and a mask.

        The original data should not be altered when a mask is input.
        """
        xc_ref = 24.7
        yc_ref = 25.2
        data = _make_gaussian_source((51, 51), 2.4, xc_ref, yc_ref, 5.0, 5.0,
                                     0)
        mask = data < 1
        xycen = centroid_quadratic(data, mask=mask)
        assert ~np.any(np.isnan(data))
        assert_allclose(xycen, (xc_ref, yc_ref), atol=0.01)

    def test_mask(self, test_data):
        """
        Test centroid_sources with mask input.
        """
        data = test_data[0]
        xcen1, ycen1 = centroid_sources(data, 25, 23, box_size=(55, 55))

        mask = np.zeros(data.shape, dtype=bool)
        mask[0, 0] = True
        mask[24, 24] = True
        mask[11, 24] = True
        xcen2, ycen2 = centroid_sources(data, 25, 23, box_size=(55, 55),
                                        mask=mask)
        assert not np.allclose(xcen1, xcen2)
        assert not np.allclose(ycen1, ycen2)

    def test_error_none(self, test_data):
        """
        Test centroid_sources with error=None for Gaussian centroids.
        """
        data = test_data[0]
        xycen1 = centroid_sources(data, xpos=25, ypos=25, error=None,
                                  centroid_func=centroid_1dg)
        xycen2 = centroid_sources(data, xpos=25, ypos=25, error=None,
                                  centroid_func=centroid_2dg)
        assert_allclose(xycen1, ([25], [25]), atol=1.0e-3)
        assert_allclose(xycen2, ([25], [25]), atol=1.0e-3)

    @pytest.mark.filterwarnings(r'ignore:.*was deprecated')
    def test_xypeaks_none(self, test_data):
        """
        Test centroid_sources with xpeak and ypeak as None for
        centroid_quadratic.
        """
        data = test_data[0]
        xycen1 = centroid_sources(data, xpos=25, ypos=25, error=None,
                                  xpeak=None, ypeak=25,
                                  centroid_func=centroid_quadratic)
        xycen2 = centroid_sources(data, xpos=25, ypos=25, error=None,
                                  xpeak=25, ypeak=None,
                                  centroid_func=centroid_quadratic)
        xycen3 = centroid_sources(data, xpos=25, ypos=25, error=None,
                                  xpeak=None, ypeak=None,
                                  centroid_func=centroid_quadratic)
        assert_allclose(xycen1, ([25], [25]), atol=1.0e-3)
        assert_allclose(xycen2, ([25], [25]), atol=1.0e-3)
        assert_allclose(xycen3, ([25], [25]), atol=1.0e-3)

    def test_centroid_quadratic_kwargs(self):
        """
        Test centroid_sources with centroid_quadratic and various
        keyword arguments.
        """
        data = np.zeros((11, 11))
        data[5, 5] = 100
        data[7, 7] = 110
        data[9, 9] = 120

        with pytest.warns(AstropyDeprecationWarning):
            xycen3 = centroid_sources(data, xpos=7, ypos=7, box_size=5,
                                      centroid_func=centroid_quadratic,
                                      xpeak=7, ypeak=7, fit_boxsize=3)
        assert_allclose(xycen3, ([7], [7]))

    def test_mask_wrong_shape(self):
        """
        Test centroid_sources raises ValueError when the mask shape
        does not match the data shape.
        """
        data = np.ones((50, 50))
        mask = np.zeros((30, 30), dtype=bool)
        match = 'mask and data must have the same shape'
        with pytest.raises(ValueError, match=match):
            centroid_sources(data, 25, 25, box_size=11, mask=mask)

    def test_xypeak_multiple_sources(self):
        """
        Test that xpeak/ypeak are correctly offset for each source in
        a multi-source centroid_sources call.
        """
        # Two isolated peaks close together so that xpeak=10 (absolute)
        # lies within both source cutouts (box_size=5):
        #   source 1 cutout: x[8:13], start=8 -> relative xpeak = 10-8 = 2
        #   source 2 cutout: x[9:14], start=9 -> relative xpeak = 10-9 = 1
        data = np.zeros((25, 25))
        data[10, 10] = 100.0
        data[10, 11] = 100.0

        with pytest.warns(AstropyDeprecationWarning):
            xc_multi, yc_multi = centroid_sources(
                data, xpos=[10, 11], ypos=[10, 10], box_size=5,
                centroid_func=centroid_quadratic,
                xpeak=10, ypeak=10, fit_boxsize=3)

        # Compare with individual single-source calls using the same
        # xpeak/ypeak to get the reference values.
        with pytest.warns(AstropyDeprecationWarning):
            xc1, yc1 = centroid_sources(
                data, xpos=10, ypos=10, box_size=5,
                centroid_func=centroid_quadratic,
                xpeak=10, ypeak=10, fit_boxsize=3)
        with pytest.warns(AstropyDeprecationWarning):
            xc2, yc2 = centroid_sources(
                data, xpos=11, ypos=10, box_size=5,
                centroid_func=centroid_quadratic,
                xpeak=10, ypeak=10, fit_boxsize=3)

        assert_allclose(xc_multi[0], xc1[0])
        assert_allclose(yc_multi[0], yc1[0])
        assert_allclose(xc_multi[1], xc2[0])
        assert_allclose(yc_multi[1], yc2[0])


def test_centroid_sources_error_multiple_sources():
    """
    Test that centroid_sources correctly applies an error array for
    multiple sources.
    """
    xpos = [25.0, 75.0]
    ypos = [30.0, 70.0]
    data1 = _make_gaussian_source((100, 100), 10.0, xpos[0], ypos[0],
                                  4.0, 4.0, 0)
    data2 = _make_gaussian_source((100, 100), 10.0, xpos[1], ypos[1],
                                  4.0, 4.0, 0)
    data = data1 + data2

    error = np.ones(data.shape, dtype=float)

    xc, yc = centroid_sources(data, xpos, ypos, box_size=21,
                              centroid_func=centroid_1dg, error=error)
    assert_allclose(xc, xpos)
    assert_allclose(yc, ypos)

    xc, yc = centroid_sources(data, xpos, ypos, box_size=21,
                              centroid_func=centroid_2dg, error=error)
    assert_allclose(xc, xpos)
    assert_allclose(yc, ypos)


def test_centroid_sources_mutation():
    """
    Test that centroid_sources does not mutate the input data or mask.
    """
    data = np.ones((50, 50))
    mask = np.zeros((50, 50), dtype=bool)
    mask[10, 10] = True
    xpos = [25.0]
    ypos = [26.0]
    data_orig = data.copy()
    mask_orig = mask.copy()
    centroid_sources(data, xpos, ypos, box_size=11, mask=mask)
    assert_array_equal(data, data_orig)
    assert_array_equal(mask, mask_orig)


def test_cutout_mask():
    """
    Test that the cutout is not completely masked (see #1514).
    """
    data = make_4gaussians_image()
    x_init = (25, 91, 151, 160)
    y_init = (40, 61, 24, 71)
    footprint = np.zeros((3, 3))
    match = 'is completely masked'
    with pytest.raises(ValueError, match=match):
        _ = centroid_sources(data, x_init, y_init, footprint=footprint,
                             centroid_func=centroid_com)

    footprint = np.zeros(data.shape, dtype=bool)
    with pytest.raises(ValueError, match=match):
        _ = centroid_sources(data, x_init, y_init, footprint=footprint,
                             centroid_func=centroid_com)

    mask = np.ones(data.shape, dtype=bool)
    with pytest.raises(ValueError, match=match):
        _ = centroid_sources(data, x_init, y_init, box_size=11, mask=mask)


class TestCentroidQuadraticClass:
    """
    Test the CentroidQuadratic class.
    """

    @pytest.mark.parametrize('x_std', [3.2, 4.0])
    @pytest.mark.parametrize('y_std', [5.7, 4.1])
    @pytest.mark.parametrize('theta', np.deg2rad([30.0, 45.0]))
    def test_basic(self, x_std, y_std, theta):
        """
        Test basic CentroidQuadratic functionality.
        """
        xc_ref = 25.7
        yc_ref = 26.2
        data = _make_gaussian_source((50, 47), 2.4, xc_ref, yc_ref,
                                     x_std, y_std, theta)

        # Test with default parameters
        centroid_func = CentroidQuadratic()
        xc, yc = centroid_func(data)
        assert_allclose((xc, yc), (xc_ref, yc_ref), rtol=0, atol=0.015)

    def test_mask(self):
        """
        Test CentroidQuadratic with mask input.
        """
        xc_ref = 25.7
        yc_ref = 26.2
        data = _make_gaussian_source((50, 47), 2.4, xc_ref, yc_ref,
                                     3.2, 5.7, 0)

        # Add an outlier
        x0 = 11
        y0 = 15
        data[y0, x0] = 1.0e5
        mask = np.zeros(data.shape, dtype=bool)
        mask[y0, x0] = True

        centroid_func = CentroidQuadratic()
        xc, yc = centroid_func(data, mask=mask)
        assert_allclose((xc, yc), (xc_ref, yc_ref), rtol=0, atol=0.015)

    def test_fit_boxsize(self):
        """
        Test CentroidQuadratic with custom fit_boxsize.
        """
        data = np.zeros((11, 11))
        data[5, 5] = 100
        data[7, 7] = 110
        data[9, 9] = 120

        centroid_func = CentroidQuadratic(fit_boxsize=3)
        xycen = centroid_func(data)
        assert_allclose(xycen, (9, 9))

    def test_with_centroid_sources(self):
        """
        Test CentroidQuadratic with centroid_sources function.
        """
        data = np.zeros((11, 11))
        data[5, 5] = 100
        data[7, 7] = 110
        data[9, 9] = 120

        # Test with custom fit_boxsize
        centroid_func = CentroidQuadratic(fit_boxsize=3)
        xycen = centroid_sources(data, xpos=5, ypos=5, box_size=7,
                                 centroid_func=centroid_func)
        assert_allclose(xycen, ([7], [7]))

    def test_repr(self):
        """
        Test CentroidQuadratic __repr__ method.
        """
        centroid_func = CentroidQuadratic()
        cls_repr = repr(centroid_func)
        assert cls_repr == 'CentroidQuadratic(fit_boxsize=5)'

        centroid_func = CentroidQuadratic(fit_boxsize=3)
        cls_repr = repr(centroid_func)
        assert cls_repr == 'CentroidQuadratic(fit_boxsize=3)'

        centroid_func = CentroidQuadratic(fit_boxsize=(3, 5))
        cls_repr = repr(centroid_func)
        assert cls_repr == 'CentroidQuadratic(fit_boxsize=(3, 5))'

    def test_str(self):
        """
        Test CentroidQuadratic __str__ method.
        """
        centroid_func = CentroidQuadratic()
        cls_str = str(centroid_func)
        cls_name = 'photutils.centroids.core.CentroidQuadratic'
        expected = f'<{cls_name}>\nfit_boxsize: 5'
        assert cls_str == expected

        centroid_func = CentroidQuadratic(fit_boxsize=3)
        cls_str = str(centroid_func)
        expected = f'<{cls_name}>\nfit_boxsize: 3'
        assert cls_str == expected



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Descriptor classes for aperture attribute validation.
"""

import astropy.units as u
import numpy as np
from astropy.coordinates import SkyCoord

__all__ = [
    'ApertureAttribute',
    'PixelPositions',
    'PositiveScalar',
    'ScalarAngle',
    'ScalarAngleOrValue',
    'SkyCoordPositions',
]


class ApertureAttribute:
    """
    Base descriptor class for aperture attribute validation.

    Parameters
    ----------
    doc : str, optional
        The description string for the attribute.
    """

    def __init__(self, doc=''):
        self.__doc__ = doc
        self.name = ''

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return instance.__dict__[self.name]

    def __set__(self, instance, value):
        self._validate(value)
        if not isinstance(value, (u.Quantity, SkyCoord)):
            value = float(value)
        # No need to reset if not already in the instance dict
        if self.name in instance.__dict__:
            self._reset_lazyproperties(instance)
        instance.__dict__[self.name] = value

    def _reset_lazyproperties(self, instance):
        # Reset lazyproperties (if they exist) for aperture parameter
        # changes
        try:
            for key in instance._lazyproperties:
                instance.__dict__.pop(key, None)
        except AttributeError:
            pass

    def __delete__(self, instance):
        del instance.__dict__[self.name]

    def _validate(self, value):
        """
        Validate the attribute value.

        An exception is raised if the value is invalid.
        """


class PixelPositions(ApertureAttribute):
    """
    Validate and set positions for pixel-based apertures.

    Pixel positions are converted to a 2D `~numpy.ndarray`.
    """

    def __set__(self, instance, value):
        # Needed for zip positions (e.g., positions = zip(xpos, ypos))
        if isinstance(value, zip):
            value = tuple(value)

        value = self._validate(value)  # np.ndarray
        # No need to reset if not already in the instance dict
        if self.name in instance.__dict__:
            self._reset_lazyproperties(instance)
        instance.__dict__[self.name] = value

    def _validate(self, value):
        try:
            value = np.asanyarray(value).astype(float)  # np.ndarray
        except TypeError as exc:
            # Value is a zip object containing Quantity objects
            msg = f'{self.name!r} must not be a Quantity'
            raise TypeError(msg) from exc

        if isinstance(value, u.Quantity):
            msg = f'{self.name!r} must not be a Quantity'
            raise TypeError(msg)

        if np.any(~np.isfinite(value)):
            msg = (f'{self.name!r} must not contain any non-finite '
                   '(e.g., NaN or inf) positions')
            raise ValueError(msg)

        value_2d = np.atleast_2d(value)
        if value_2d.ndim > 2 or value_2d.shape[1] != 2:
            msg = (f'{self.name!r} must be a (x, y) pixel position '
                   'or a list or array of (x, y) pixel positions, '
                   'e.g., [(x1, y1), (x2, y2), (x3, y3)]')
            raise ValueError(msg)

        return value


class SkyCoordPositions(ApertureAttribute):
    """
    Check that value is a `~astropy.coordinates.SkyCoord`.
    """

    def _validate(self, value):
        if not isinstance(value, SkyCoord):
            msg = f'{self.name!r} must be a SkyCoord instance'
            raise TypeError(msg)


class PositiveScalar(ApertureAttribute):
    """
    Check that value is a strictly positive (> 0) scalar.
    """

    def _validate(self, value):
        if not np.isscalar(value) or value <= 0:
            msg = f'{self.name!r} must be a positive scalar'
            raise ValueError(msg)


class ScalarAngle(ApertureAttribute):
    """
    Check that value is a scalar angle, either as a
    `~astropy.coordinates.Angle` or `~astropy.units.Quantity` with
    angular units.
    """

    def _validate(self, value):
        if isinstance(value, u.Quantity):
            if not value.isscalar:
                msg = f'{self.name!r} must be a scalar'
                raise ValueError(msg)

            if value.unit.physical_type != 'angle':
                msg = f'{self.name!r} must have angular units'
                raise ValueError(msg)
        else:
            msg = f'{self.name!r} must be a scalar angle'
            raise TypeError(msg)


class PositiveScalarAngle(ApertureAttribute):
    """
    Check that value is a positive scalar angle, either as a
    `~astropy.coordinates.Angle` or `~astropy.units.Quantity` with
    angular units.
    """

    def _validate(self, value):
        if isinstance(value, u.Quantity):
            if not value.isscalar:
                msg = f'{self.name!r} must be a scalar'
                raise ValueError(msg)

            if value.unit.physical_type != 'angle':
                msg = f'{self.name!r} must have angular units'
                raise ValueError(msg)
        else:
            msg = f'{self.name!r} must be a scalar angle'
            raise TypeError(msg)

        if value <= 0:
            msg = f'{self.name!r} must be greater than zero'
            raise ValueError(msg)


class ScalarAngleOrValue(ApertureAttribute):
    """
    Check that value is a scalar angle, either as a
    `~astropy.coordinates.Angle` or `~astropy.units.Quantity` with
    angular units, or a scalar float.

    The value is always output as a `~astropy.units.Quantity` with
    angular units. If the value is not a `~astropy.units.Quantity`, it
    is assumed to be in radians.
    """

    def __set__(self, instance, value):
        self._validate(value)
        # No need to reset if not already in the instance dict
        if self.name in instance.__dict__:
            self._reset_lazyproperties(instance)

        # If theta is not a Quantity, it is assumed to be in radians
        if not isinstance(value, u.Quantity):
            value <<= u.radian
        instance.__dict__[self.name] = value

    def _validate(self, value):
        if isinstance(value, u.Quantity):
            if not value.isscalar:
                msg = f'{self.name!r} must be a scalar'
                raise ValueError(msg)

            if value.unit.physical_type != 'angle':
                msg = f'{self.name!r} must have angular units'
                raise ValueError(msg)
        elif not np.isscalar(value):
            msg = (f'If not an angle Quantity, {self.name!r} must be a '
                   'scalar float in radians')
            raise ValueError(msg)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the sample module.
"""

import numpy as np
import pytest

from photutils.isophote.integrator import (BILINEAR, MEAN, MEDIAN,
                                           NEAREST_NEIGHBOR)
from photutils.isophote.isophote import Isophote
from photutils.isophote.sample import EllipseSample
from photutils.isophote.tests.make_test_data import make_test_image

DEFAULT_FIX = np.array([False, False, False, False])

DATA = make_test_image(background=100.0, i0=0.0, noise=10.0, seed=0)


# the median is not so good at estimating rms
@pytest.mark.parametrize(('integrmode', 'amin', 'amax'),
                         [(NEAREST_NEIGHBOR, 7.0, 15.0),
                          (BILINEAR, 7.0, 15.0),
                          (MEAN, 7.0, 15.0),
                          (MEDIAN, 6.0, 15.0)])
def test_scatter(integrmode, amin, amax):
    """
    Check that the pixel standard deviation can be reliably estimated
    from the rms scatter and the sector area.

    The test data is just a flat image with noise, no galaxy. We define
    the noise rms and then compare how close the pixel std dev estimated
    at extraction matches this input noise.
    """
    sample = EllipseSample(DATA, 50.0, astep=0.2, integrmode=integrmode)
    sample.update(fixed_parameters=DEFAULT_FIX)
    iso = Isophote(sample, 0, valid=True, stop_code=0)

    assert iso.pix_stddev < amax
    assert iso.pix_stddev > amin


def test_coordinates():
    sample = EllipseSample(DATA, 50.0)
    sample.update(fixed_parameters=DEFAULT_FIX)
    x, y = sample.coordinates()

    assert isinstance(x, np.ndarray)
    assert isinstance(y, np.ndarray)


def test_sclip():
    sample = EllipseSample(DATA, 50.0, n_clip=3)
    sample.update(fixed_parameters=DEFAULT_FIX)
    x, y = sample.coordinates()

    assert isinstance(x, np.ndarray)
    assert isinstance(y, np.ndarray)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Configuration file for the pytest test suite.
"""

from importlib.metadata import packages_distributions, requires, version

from packaging.requirements import Requirement

from photutils.utils._optional_deps import _DIST_TO_IMPORT

try:
    from pytest_astropy_header.display import (PYTEST_HEADER_MODULES,
                                               TESTED_VERSIONS)
except ImportError:
    PYTEST_HEADER_MODULES = {}
    TESTED_VERSIONS = {}


def pytest_configure():
    """
    Configure pytest settings.
    """
    # Resolve the distribution name for this package
    import_name = __package__
    dist_name = packages_distributions().get(import_name, [import_name])[0]

    # Collect all dependency import names (core + 'all' extra)
    dep_names = set()
    for req_str in (requires(dist_name) or []):
        req = Requirement(req_str)
        if not req.marker or req.marker.evaluate({'extra': 'all'}):
            dep_names.add(_DIST_TO_IMPORT.get(req.name, req.name))

    PYTEST_HEADER_MODULES.clear()
    for dep in sorted(dep_names):
        PYTEST_HEADER_MODULES[dep] = dep

    TESTED_VERSIONS[dist_name] = version(dist_name)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the ellipse module.
"""

import math

import numpy as np
import pytest
from astropy.io import fits
from astropy.modeling.models import Gaussian2D
from astropy.utils.exceptions import AstropyUserWarning

from photutils.datasets import make_noise_image
from photutils.datasets.load import _get_path
from photutils.isophote.ellipse import Ellipse
from photutils.isophote.geometry import EllipseGeometry
from photutils.isophote.isophote import Isophote, IsophoteList
from photutils.isophote.tests.make_test_data import make_test_image

# define an off-center position and a tilted sma
POS = 384
PA = np.deg2rad(10.0)

# build off-center test data. It's fine to have a single np array to use
# in all tests that need it, but do not use a single instance of
# EllipseGeometry. The code may eventually modify it's contents. The safe
# bet is to build it wherever it's needed. The cost is negligible.
OFFSET_GALAXY = make_test_image(x0=POS, y0=POS, pa=PA, noise=1.0e-12,
                                seed=0)


class TestEllipse:
    def setup_class(self):
        # centered, tilted galaxy
        self.data = make_test_image(pa=PA, seed=0)

    @pytest.mark.remote_data
    def test_find_center(self):
        path = _get_path('isophote/M51.fits', location='photutils-datasets',
                         cache=True)
        hdu = fits.open(path)
        data = hdu[0].data
        hdu.close()

        geometry = EllipseGeometry(252, 253, 10.0, 0.2, np.pi / 2)
        geometry.find_center(data)
        assert geometry.x0 == 257.0
        assert geometry.y0 == 258.0

    def test_basic(self):
        ellipse = Ellipse(self.data)
        isophote_list = ellipse.fit_image()

        assert isinstance(isophote_list, IsophoteList)
        assert len(isophote_list) > 1
        assert isinstance(isophote_list[0], Isophote)

        # verify that the list is properly sorted in sem-major axis length
        assert isophote_list[-1] > isophote_list[0]

        # the fit should stop where gradient loses reliability.
        assert len(isophote_list) == 69
        assert isophote_list[-1].stop_code == 1

    def test_linear(self):
        ellipse = Ellipse(self.data)
        isophote_list = ellipse.fit_image(linear=True, step=2.0)

        # verify that the list is properly sorted in sem-major axis length
        assert isophote_list[-1] > isophote_list[0]

        # difference in sma between successive isohpotes must be constant.
        step = isophote_list[-1].sma - isophote_list[-2].sma
        assert math.isclose((isophote_list[-2].sma - isophote_list[-3].sma),
                            step, rel_tol=0.01)
        assert math.isclose((isophote_list[-3].sma - isophote_list[-4].sma),
                            step, rel_tol=0.01)
        assert math.isclose((isophote_list[2].sma - isophote_list[1].sma),
                            step, rel_tol=0.01)

    def test_fit_one_ellipse(self):
        ellipse = Ellipse(self.data)
        isophote = ellipse.fit_isophote(40.0)

        assert isinstance(isophote, Isophote)
        assert isophote.valid

    def test_offcenter_fail(self):
        # A first guess ellipse that is centered in the image frame.
        # This should result in failure since the real galaxy
        # image is off-center by a large offset.
        ellipse = Ellipse(OFFSET_GALAXY)

        match1 = 'Degrees of freedom'
        match2 = 'Mean of empty slice'
        match3 = 'invalid value encountered'
        match4 = 'No meaningful fit was possible'
        ctx1 = pytest.warns(RuntimeWarning, match=match1)
        ctx2 = pytest.warns(RuntimeWarning, match=match2)
        ctx3 = pytest.warns(RuntimeWarning, match=match3)
        ctx4 = pytest.warns(AstropyUserWarning, match=match4)
        with ctx1, ctx2, ctx3, ctx4:
            isophote_list = ellipse.fit_image()
            assert len(isophote_list) == 0

    def test_offcenter_fit(self):
        # A first guess ellipse that is roughly centered on the
        # offset galaxy image.
        g = EllipseGeometry(POS + 5, POS + 5, 10.0, eps=0.2, pa=PA, astep=0.1)
        ellipse = Ellipse(OFFSET_GALAXY, geometry=g)
        isophote_list = ellipse.fit_image()

        # the fit should stop when too many potential sample
        # points fall outside the image frame.
        assert len(isophote_list) == 63
        assert isophote_list[-1].stop_code == 1

    def test_offcenter_go_beyond_frame(self):
        # Same as before, but now force the fit to goo
        # beyond the image frame limits.
        g = EllipseGeometry(POS + 5, POS + 5, 10.0, eps=0.2, pa=PA, astep=0.1)
        ellipse = Ellipse(OFFSET_GALAXY, geometry=g)
        isophote_list = ellipse.fit_image(maxsma=400.0)

        # the fit should go to maxsma, but with fixed geometry
        assert len(isophote_list) == 71
        assert isophote_list[-1].stop_code == 4

        # check that no zero-valued intensities were left behind
        # in the sample arrays when sampling outside the image.
        for iso in isophote_list:
            assert not np.any(iso.sample.values[2] == 0)

    def test_ellipse_shape(self):
        """
        Regression test for #670/673.
        """
        ny = 500
        nx = 150
        g = Gaussian2D(100.0, nx / 2.0, ny / 2.0, 20, 12,
                       theta=np.deg2rad(40.0))
        y, x = np.mgrid[0:ny, 0:nx]
        noise = make_noise_image((ny, nx), distribution='gaussian', mean=0.0,
                                 stddev=2.0, seed=0)
        data = g(x, y) + noise
        ellipse = Ellipse(data)  # estimates initial center
        isolist = ellipse.fit_image()
        assert len(isolist) == 54



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Rectangular and rectangular-annulus apertures in both pixel and sky
coordinates.
"""

import math

import astropy.units as u
import numpy as np
from astropy.coordinates import Angle
from astropy.utils import lazyproperty

from photutils.aperture.attributes import (PixelPositions, PositiveScalar,
                                           PositiveScalarAngle, ScalarAngle,
                                           ScalarAngleOrValue,
                                           SkyCoordPositions)
from photutils.aperture.core import PixelAperture, SkyAperture
from photutils.aperture.mask import ApertureMask
from photutils.geometry import rectangular_overlap_grid
from photutils.utils._deprecation import (deprecated,
                                          deprecated_positional_kwargs)
from photutils.utils._wcs_helpers import (pixel_to_sky_scales,
                                          sky_to_pixel_scales)

__all__ = [
    'RectangularAnnulus',
    'RectangularAperture',
    'RectangularMaskMixin',
    'SkyRectangularAnnulus',
    'SkyRectangularAperture',
]


@deprecated('3.0', until='4.0')
class RectangularMaskMixin:  # pragma: no cover
    """
    Mixin class to create masks for rectangular or rectangular-annulus
    aperture objects.

    .. deprecated:: 3.0
    """

    def to_mask(self, method='exact', subpixels=5):
        """
        Return a mask for the aperture.

        Parameters
        ----------
        method : {'exact', 'center', 'subpixel'}, optional
            The method used to determine the overlap of the aperture
            on the pixel grid. Not all options are available for all
            aperture types. Note that the more precise methods are
            generally slower. The following methods are available:

            * ``'exact'`` (default):
              The exact fractional overlap of the aperture and each
              pixel is calculated. The aperture weights will contain
              values between 0 and 1.

            * ``'center'``:
              A pixel is considered to be entirely in or out of the
              aperture depending on whether its center is in or out of
              the aperture. The aperture weights will contain values
              only of 0 (out) and 1 (in).

            * ``'subpixel'``:
              A pixel is divided into subpixels (see the ``subpixels``
              keyword), each of which are considered to be entirely in
              or out of the aperture depending on whether its center is
              in or out of the aperture. If ``subpixels=1``, this method
              is equivalent to ``'center'``. The aperture weights will
              contain values between 0 and 1.

        subpixels : int, optional
            For the ``'subpixel'`` method, resample pixels by this
            factor in each dimension. That is, each pixel is divided
            into ``subpixels**2`` subpixels. This keyword is ignored
            unless ``method='subpixel'``.

        Returns
        -------
        mask : `~photutils.aperture.ApertureMask` or list of \
                `~photutils.aperture.ApertureMask`
            A mask for the aperture. If the aperture is scalar then
            a single `~photutils.aperture.ApertureMask` is returned,
            otherwise a list of `~photutils.aperture.ApertureMask` is
            returned.
        """
        _, subpixels = self._translate_mask_method(method, subpixels,
                                                   rectangle=True)

        if hasattr(self, 'w'):
            w = self.w
            h = self.h
        elif hasattr(self, 'w_out'):  # annulus
            w = self.w_out
            h = self.h_out
        else:
            msg = 'Cannot determine the aperture radius'
            raise ValueError(msg)

        masks = []
        for bbox, edges in zip(self._bbox, self._centered_edges, strict=True):
            ny, nx = bbox.shape
            theta_rad = self.theta.to(u.radian).value
            mask = rectangular_overlap_grid(edges[0], edges[1], edges[2],
                                            edges[3], nx, ny, w, h,
                                            theta_rad, 0, subpixels)

            # Subtract the inner rectangle for an annulus
            if hasattr(self, 'w_in'):
                mask -= rectangular_overlap_grid(edges[0], edges[1], edges[2],
                                                 edges[3], nx, ny, self.w_in,
                                                 self.h_in, theta_rad,
                                                 0, subpixels)

            masks.append(ApertureMask(mask, bbox))

        if self.isscalar:
            return masks[0]

        return masks

    @staticmethod
    def _calc_extents(width, height, theta):
        """
        Calculate half of the bounding box extents of a rectangle.
        """
        return _calc_rectangle_extents(width, height, theta)

    @staticmethod
    def _lower_left_positions(positions, width, height, theta):
        """
        Calculate lower-left positions from the input center positions.

        Used for creating `~matplotlib.patches.Rectangle` patch for the
        aperture.
        """
        return _calc_lower_left_positions(positions, width, height, theta)


def _calc_rectangle_extents(width, height, theta):
    """
    Calculate half of the bounding box extents of a rectangle.
    """
    theta_rad = theta.to(u.radian).value
    half_width = width / 2.0
    half_height = height / 2.0
    sin_theta = math.sin(theta_rad)
    cos_theta = math.cos(theta_rad)
    x_extent1 = abs((half_width * cos_theta) - (half_height * sin_theta))
    x_extent2 = abs((half_width * cos_theta) + (half_height * sin_theta))
    y_extent1 = abs((half_width * sin_theta) + (half_height * cos_theta))
    y_extent2 = abs((half_width * sin_theta) - (half_height * cos_theta))
    x_extent = max(x_extent1, x_extent2)
    y_extent = max(y_extent1, y_extent2)

    return x_extent, y_extent


def _calc_lower_left_positions(positions, width, height, theta):
    """
    Calculate lower-left positions from the input center positions.

    Used for creating `~matplotlib.patches.Rectangle` patch for the
    aperture.
    """
    theta_rad = theta.to(u.radian).value
    half_width = width / 2.0
    half_height = height / 2.0
    sin_theta = math.sin(theta_rad)
    cos_theta = math.cos(theta_rad)
    xshift = (half_height * sin_theta) - (half_width * cos_theta)
    yshift = -(half_height * cos_theta) - (half_width * sin_theta)

    return np.atleast_2d(positions) + np.array([xshift, yshift])


class RectangularAperture(PixelAperture):
    """
    A rectangular aperture defined in pixel coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : array_like
        The pixel coordinates of the aperture center(s) in one of the
        following formats:

        * single ``(x, y)`` pair as a tuple, list, or `~numpy.ndarray`
        * tuple, list, or `~numpy.ndarray` of ``(x, y)`` pairs

    w : float
        The full width of the rectangle in pixels. For ``theta=0`` the
        width side is along the ``x`` axis.

    h : float
        The full height of the rectangle in pixels. For ``theta=0`` the
        height side is along the ``y`` axis.

    theta : float or `~astropy.units.Quantity`, optional
        The rotation angle as an angular quantity
        (`~astropy.units.Quantity` or `~astropy.coordinates.Angle`) or
        value in radians (as a float) from the positive ``x`` axis. The
        rotation angle increases counterclockwise.

    Raises
    ------
    ValueError : `ValueError`
        If either width (``w``) or height (``h``) is negative.

    Examples
    --------
    >>> from astropy.coordinates import Angle
    >>> from photutils.aperture import RectangularAperture

    >>> theta = Angle(80, 'deg')
    >>> aper = RectangularAperture([10.0, 20.0], 5.0, 3.0)
    >>> aper = RectangularAperture((10.0, 20.0), 5.0, 3.0, theta=theta)

    >>> pos1 = (10.0, 20.0)  # (x, y)
    >>> pos2 = (30.0, 40.0)
    >>> pos3 = (50.0, 60.0)
    >>> aper = RectangularAperture([pos1, pos2, pos3], 5.0, 3.0)
    >>> aper = RectangularAperture((pos1, pos2, pos3), 5.0, 3.0, theta=theta)
    """

    _params = ('positions', 'w', 'h', 'theta')
    positions = PixelPositions('The center pixel position(s).')
    w = PositiveScalar('The full width in pixels.')
    h = PositiveScalar('The full height in pixels.')
    theta = ScalarAngleOrValue('The counterclockwise rotation angle as an '
                               'angular Quantity or a value in radians from '
                               'the positive x axis.')
    _is_rectangle = True  # remove when rectangles support "exact" method

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, positions, w, h, theta=0.0):
        self.positions = positions
        self.w = w
        self.h = h
        self.theta = theta

    @lazyproperty
    def _xy_extents(self):
        """
        The half-width and half-height of the bounding box of the
        rectangle.
        """
        return _calc_rectangle_extents(self.w, self.h, self.theta)

    @lazyproperty
    def area(self):
        """
        The exact geometric area of the aperture shape.
        """
        return self.w * self.h

    def _to_patch(self, *, origin=(0, 0), **kwargs):
        """
        Return a `~matplotlib.patches.Patch` for the aperture.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` position of the origin of the displayed
            image.

        **kwargs : dict, optional
            Any keyword arguments accepted by
            `matplotlib.patches.Patch`.

        Returns
        -------
        patch : `~matplotlib.patches.Patch` or list of \
                `~matplotlib.patches.Patch`
            A patch for the aperture. If the aperture is scalar then a
            single `~matplotlib.patches.Patch` is returned, otherwise a
            list of `~matplotlib.patches.Patch` is returned.
        """
        import matplotlib.patches as mpatches

        xy_positions, patch_kwargs = self._define_patch_params(origin=origin,
                                                               **kwargs)
        xy_positions = _calc_lower_left_positions(xy_positions, self.w,
                                                  self.h, self.theta)

        angle = self.theta.to(u.deg).value
        patches = [mpatches.Rectangle(xy_position, self.w, self.h,
                                      angle=angle, **patch_kwargs)
                   for xy_position in xy_positions]

        if self.isscalar:
            return patches[0]

        return patches

    def _compute_overlap(self, edges, nx, ny, use_exact, subpixels):
        """
        Compute the overlap of the aperture on the pixel grid.

        Parameters
        ----------
        edges : list of 4 1D `~numpy.ndarray`
            The edges of the pixel grid in the form of
            ``[x_edges, y_edges, x_centers, y_centers]``.

        nx, ny : int
            The number of pixels in the x and y directions.

        use_exact : bool
            Whether to use the exact method for calculating the overlap.

        subpixels : int
            The number of subpixels to use in each dimension for the
            subpixel method.

        Returns
        -------
        overlap : 2D `~numpy.ndarray`
            The overlap of the aperture on the pixel grid. The values
            will be between 0 and 1, where 0 means no overlap and 1
            means full overlap.
        """
        theta_rad = self.theta.to(u.radian).value
        return rectangular_overlap_grid(edges[0], edges[1], edges[2],
                                        edges[3], nx, ny, self.w,
                                        self.h, theta_rad,
                                        use_exact, subpixels)

    def to_sky(self, wcs):
        """
        Convert the aperture to a `SkyRectangularAperture` object
        defined in celestial coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `SkyRectangularAperture` object
            A `SkyRectangularAperture` object.

        Notes
        -----
        The aperture shape parameters are converted using the local WCS
        properties (pixel scale, rotation angle) evaluated at the first
        aperture position. Because aperture objects require scalar shape
        parameters, only a single reference position is used for the
        conversion. For apertures with multiple positions used with a
        WCS that has spatially-varying distortions, this may produce
        inaccurate results for positions far from the first position.
        """
        xpos, ypos = np.transpose(self.positions)
        positions = wcs.pixel_to_world(xpos, ypos)

        first_pos = np.atleast_2d(self.positions)[0]
        pixcoord = (float(first_pos[0]), float(first_pos[1]))
        _, scale_w, scale_h, sky_angle = pixel_to_sky_scales(
            pixcoord, wcs, self.theta.to(u.rad).value)

        w = Angle(self.w * scale_w, 'arcsec')
        h = Angle(self.h * scale_h, 'arcsec')
        return SkyRectangularAperture(positions=positions, w=w, h=h,
                                      theta=sky_angle)


class RectangularAnnulus(PixelAperture):
    r"""
    A rectangular annulus aperture defined in pixel coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : array_like
        The pixel coordinates of the aperture center(s) in one of the
        following formats:

        * single ``(x, y)`` pair as a tuple, list, or `~numpy.ndarray`
        * tuple, list, or `~numpy.ndarray` of ``(x, y)`` pairs

    w_in : float
        The inner full width of the rectangular annulus in pixels. For
        ``theta=0`` the width side is along the ``x`` axis.

    w_out : float
        The outer full width of the rectangular annulus in pixels. For
        ``theta=0`` the width side is along the ``x`` axis.

    h_out : float
        The outer full height of the rectangular annulus in pixels.

    h_in : `None` or float
        The inner full height of the rectangular annulus in pixels. If
        `None`, then the inner full height is calculated as:

        .. math::

            h_{in} = h_{out} \left(\frac{w_{in}}{w_{out}}\right)

        For ``theta=0`` the height side is along the ``y`` axis.

    theta : float or `~astropy.units.Quantity`, optional
        The rotation angle as an angular quantity
        (`~astropy.units.Quantity` or `~astropy.coordinates.Angle`) or
        value in radians (as a float) from the positive ``x`` axis. The
        rotation angle increases counterclockwise.

    Raises
    ------
    ValueError : `ValueError`
        If inner width (``w_in``) is greater than outer width
        (``w_out``).

    ValueError : `ValueError`
        If either the inner width (``w_in``) or the outer height
        (``h_out``) is negative.

    Examples
    --------
    >>> from astropy.coordinates import Angle
    >>> from photutils.aperture import RectangularAnnulus

    >>> theta = Angle(80, 'deg')
    >>> aper = RectangularAnnulus([10.0, 20.0], 3.0, 8.0, 5.0)
    >>> aper = RectangularAnnulus((10.0, 20.0), 3.0, 8.0, 5.0, theta=theta)

    >>> pos1 = (10.0, 20.0)  # (x, y)
    >>> pos2 = (30.0, 40.0)
    >>> pos3 = (50.0, 60.0)
    >>> aper = RectangularAnnulus([pos1, pos2, pos3], 3.0, 8.0, 5.0)
    >>> aper = RectangularAnnulus((pos1, pos2, pos3), 3.0, 8.0, 5.0,
    ...                           theta=theta)
    """

    _params = ('positions', 'w_in', 'w_out', 'h_in', 'h_out', 'theta')
    positions = PixelPositions('The center pixel position(s).')
    w_in = PositiveScalar('The inner full width in pixels.')
    w_out = PositiveScalar('The outer full width in pixels.')
    h_in = PositiveScalar('The inner full height in pixels.')
    h_out = PositiveScalar('The outer full height in pixels.')
    theta = ScalarAngleOrValue('The counterclockwise rotation angle as an '
                               'angular Quantity or a value in radians from '
                               'the positive x axis.')
    _is_rectangle = True  # remove when rectangles support "exact" method

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, positions, w_in, w_out, h_out, h_in=None, theta=0.0):
        if not w_out > w_in:
            msg = "'w_out' must be greater than 'w_in'"
            raise ValueError(msg)

        self.positions = positions
        self.w_in = w_in
        self.w_out = w_out
        self.h_out = h_out

        if h_in is None:
            h_in = self.w_in * self.h_out / self.w_out
        elif not h_out > h_in:
            msg = "'h_out' must be greater than 'h_in'"
            raise ValueError(msg)
        self.h_in = h_in

        self.theta = theta

    @lazyproperty
    def _xy_extents(self):
        """
        The half-width and half-height of the bounding box of the
        rectangle.
        """
        return _calc_rectangle_extents(self.w_out, self.h_out, self.theta)

    @lazyproperty
    def area(self):
        """
        The exact geometric area of the aperture shape.
        """
        return self.w_out * self.h_out - self.w_in * self.h_in

    def _to_patch(self, *, origin=(0, 0), **kwargs):
        """
        Return a `~matplotlib.patches.Patch` for the aperture.

        Parameters
        ----------
        origin : array_like, optional
            The ``(x, y)`` position of the origin of the displayed
            image.

        **kwargs : dict, optional
            Any keyword arguments accepted by
            `matplotlib.patches.Patch`.

        Returns
        -------
        patch : `~matplotlib.patches.Patch` or list of \
                `~matplotlib.patches.Patch`
            A patch for the aperture. If the aperture is scalar then a
            single `~matplotlib.patches.Patch` is returned, otherwise a
            list of `~matplotlib.patches.Patch` is returned.
        """
        import matplotlib.patches as mpatches

        xy_positions, patch_kwargs = self._define_patch_params(origin=origin,
                                                               **kwargs)
        inner_xy_positions = _calc_lower_left_positions(xy_positions,
                                                        self.w_in,
                                                        self.h_in,
                                                        self.theta)
        outer_xy_positions = _calc_lower_left_positions(xy_positions,
                                                        self.w_out,
                                                        self.h_out,
                                                        self.theta)

        patches = []
        angle = self.theta.to(u.deg).value
        for xy_in, xy_out in zip(inner_xy_positions, outer_xy_positions,
                                 strict=True):
            patch_inner = mpatches.Rectangle(xy_in, self.w_in, self.h_in,
                                             angle=angle)
            patch_outer = mpatches.Rectangle(xy_out, self.w_out, self.h_out,
                                             angle=angle)
            path = self._make_annulus_path(patch_inner, patch_outer)
            patches.append(mpatches.PathPatch(path, **patch_kwargs))

        if self.isscalar:
            return patches[0]

        return patches

    def _compute_overlap(self, edges, nx, ny, use_exact, subpixels):
        """
        Compute the overlap of the aperture on the pixel grid.

        Parameters
        ----------
        edges : list of 4 1D `~numpy.ndarray`
            The edges of the pixel grid in the form of
            ``[x_edges, y_edges, x_centers, y_centers]``.

        nx, ny : int
            The number of pixels in the x and y directions.

        use_exact : bool
            Whether to use the exact method for calculating the overlap.

        subpixels : int
            The number of subpixels to use in each dimension for the
            subpixel method.

        Returns
        -------
        overlap : 2D `~numpy.ndarray`
            The overlap of the aperture on the pixel grid. The values
            will be between 0 and 1, where 0 means no overlap and 1
            means full overlap.
        """
        theta_rad = self.theta.to(u.radian).value
        overlap = rectangular_overlap_grid(edges[0], edges[1], edges[2],
                                           edges[3], nx, ny, self.w_out,
                                           self.h_out, theta_rad,
                                           use_exact, subpixels)
        overlap -= rectangular_overlap_grid(edges[0], edges[1], edges[2],
                                            edges[3], nx, ny, self.w_in,
                                            self.h_in, theta_rad,
                                            use_exact, subpixels)
        return overlap

    def to_sky(self, wcs):
        """
        Convert the aperture to a `SkyRectangularAnnulus` object defined
        in celestial coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `SkyRectangularAnnulus` object
            A `SkyRectangularAnnulus` object.

        Notes
        -----
        The aperture shape parameters are converted using the local WCS
        properties (pixel scale, rotation angle) evaluated at the first
        aperture position. Because aperture objects require scalar shape
        parameters, only a single reference position is used for the
        conversion. For apertures with multiple positions used with a
        WCS that has spatially-varying distortions, this may produce
        inaccurate results for positions far from the first position.
        """
        xpos, ypos = np.transpose(self.positions)
        positions = wcs.pixel_to_world(xpos, ypos)

        first_pos = np.atleast_2d(self.positions)[0]
        pixcoord = (float(first_pos[0]), float(first_pos[1]))
        _, scale_w, scale_h, sky_angle = pixel_to_sky_scales(
            pixcoord, wcs, self.theta.to(u.rad).value)

        w_in = Angle(self.w_in * scale_w, 'arcsec')
        w_out = Angle(self.w_out * scale_w, 'arcsec')
        h_in = Angle(self.h_in * scale_h, 'arcsec')
        h_out = Angle(self.h_out * scale_h, 'arcsec')
        return SkyRectangularAnnulus(positions=positions, w_in=w_in,
                                     w_out=w_out, h_out=h_out,
                                     h_in=h_in, theta=sky_angle)


class SkyRectangularAperture(SkyAperture):
    """
    A rectangular aperture defined in sky coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : `~astropy.coordinates.SkyCoord`
        The celestial coordinates of the aperture center(s). This can be
        either scalar coordinates or an array of coordinates.

    w : scalar `~astropy.units.Quantity`
        The full width of the rectangle in angular units. For
        ``theta=0`` the width side is along the North-South axis.

    h : scalar `~astropy.units.Quantity`
        The full height of the rectangle in angular units. For
        ``theta=0`` the height side is along the East-West axis.

    theta : scalar `~astropy.units.Quantity`, optional
        The position angle (in angular units) of the rectangle "width"
        side. For a right-handed world coordinate system, the position
        angle increases counterclockwise from North (PA=0).

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u
    >>> from photutils.aperture import SkyRectangularAperture
    >>> positions = SkyCoord(ra=[10.0, 20.0], dec=[30.0, 40.0], unit='deg')
    >>> aper = SkyRectangularAperture(positions, 1.0*u.arcsec, 0.5*u.arcsec)
    """

    _params = ('positions', 'w', 'h', 'theta')
    positions = SkyCoordPositions('The center position(s) in sky coordinates.')
    w = PositiveScalarAngle('The full width in angular units.')
    h = PositiveScalarAngle('The full height in angular units.')
    theta = ScalarAngle('The position angle (in angular units) of the '
                        'rectangle "width" side.')

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, positions, w, h, theta=0.0 * u.deg):
        self.positions = positions
        self.w = w
        self.h = h
        self.theta = theta

    def to_pixel(self, wcs):
        """
        Convert the aperture to a `RectangularAperture` object defined
        in pixel coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `RectangularAperture` object
            A `RectangularAperture` object.

        Notes
        -----
        The aperture shape parameters are converted using the local WCS
        properties (pixel scale, rotation angle) evaluated at the first
        aperture position. Because aperture objects require scalar shape
        parameters, only a single reference position is used for the
        conversion. For apertures with multiple positions used with a
        WCS that has spatially-varying distortions, this may produce
        inaccurate results for positions far from the first position.
        """
        xpos, ypos = wcs.world_to_pixel(self.positions)
        positions = np.transpose((xpos, ypos))

        skypos = self.positions if self.isscalar else self.positions[0]
        sky_angle_rad = self.theta.to(u.rad).value
        _, scale_w, scale_h, pixel_angle = sky_to_pixel_scales(
            skypos, wcs, sky_angle_rad)

        w = self.w.to(u.arcsec).value * scale_w
        h = self.h.to(u.arcsec).value * scale_h
        return RectangularAperture(positions=positions, w=w, h=h,
                                   theta=pixel_angle)


class SkyRectangularAnnulus(SkyAperture):
    r"""
    A rectangular annulus aperture defined in sky coordinates.

    The aperture has a single fixed size/shape, but it can have multiple
    positions (see the ``positions`` input).

    Parameters
    ----------
    positions : `~astropy.coordinates.SkyCoord`
        The celestial coordinates of the aperture center(s). This can be
        either scalar coordinates or an array of coordinates.

    w_in : scalar `~astropy.units.Quantity`
        The inner full width of the rectangular annulus in angular
        units. For ``theta=0`` the width side is along the North-South
        axis.

    w_out : scalar `~astropy.units.Quantity`
        The outer full width of the rectangular annulus in angular
        units. For ``theta=0`` the width side is along the North-South
        axis.

    h_out : scalar `~astropy.units.Quantity`
        The outer full height of the rectangular annulus in angular
        units.

    h_in : `None` or scalar `~astropy.units.Quantity`
        The inner full height of the rectangular annulus in angular
        units. If `None`, then the inner full height is calculated as:

        .. math::

            h_{in} = h_{out} \left(\frac{w_{in}}{w_{out}}\right)

        For ``theta=0`` the height side is along the East-West axis.

    theta : scalar `~astropy.units.Quantity`, optional
        The position angle (in angular units) of the rectangle "width"
        side. For a right-handed world coordinate system, the position
        angle increases counterclockwise from North (PA=0).

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord
    >>> import astropy.units as u
    >>> from photutils.aperture import SkyRectangularAnnulus
    >>> positions = SkyCoord(ra=[10.0, 20.0], dec=[30.0, 40.0], unit='deg')
    >>> aper = SkyRectangularAnnulus(positions, 3.0*u.arcsec, 8.0*u.arcsec,
    ...                              5.0*u.arcsec)
    """

    _params = ('positions', 'w_in', 'w_out', 'h_in', 'h_out', 'theta')
    positions = SkyCoordPositions('The center position(s) in sky coordinates.')
    w_in = PositiveScalarAngle('The inner full width in angular units.')
    w_out = PositiveScalarAngle('The outer full width in angular units.')
    h_in = PositiveScalarAngle('The inner full height in angular units.')
    h_out = PositiveScalarAngle('The outer full height in angular units.')
    theta = ScalarAngle('The position angle (in angular units) of the '
                        'rectangle "width" side.')

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, positions, w_in, w_out, h_out, h_in=None,
                 theta=0.0 * u.deg):
        if not w_out > w_in:
            msg = "'w_out' must be greater than 'w_in'"
            raise ValueError(msg)

        self.positions = positions
        self.w_in = w_in
        self.w_out = w_out
        self.h_out = h_out

        if h_in is None:
            h_in = self.w_in * self.h_out / self.w_out
        elif not h_out > h_in:
            msg = "'h_out' must be greater than 'h_in'"
            raise ValueError(msg)
        self.h_in = h_in

        self.theta = theta

    def to_pixel(self, wcs):
        """
        Convert the aperture to a `RectangularAnnulus` object defined in
        pixel coordinates.

        Parameters
        ----------
        wcs : WCS object
            A world coordinate system (WCS) transformation that
            supports the `astropy shared interface for WCS
            <https://docs.astropy.org/en/stable/wcs/wcsapi.html>`_
            (e.g., `astropy.wcs.WCS`, `gwcs.wcs.WCS`).

        Returns
        -------
        aperture : `RectangularAnnulus` object
            A `RectangularAnnulus` object.

        Notes
        -----
        The aperture shape parameters are converted using the local WCS
        properties (pixel scale, rotation angle) evaluated at the first
        aperture position. Because aperture objects require scalar shape
        parameters, only a single reference position is used for the
        conversion. For apertures with multiple positions used with a
        WCS that has spatially-varying distortions, this may produce
        inaccurate results for positions far from the first position.
        """
        xpos, ypos = wcs.world_to_pixel(self.positions)
        positions = np.transpose((xpos, ypos))

        skypos = self.positions if self.isscalar else self.positions[0]
        sky_angle_rad = self.theta.to(u.rad).value
        _, scale_w, scale_h, pixel_angle = sky_to_pixel_scales(
            skypos, wcs, sky_angle_rad)

        w_in = self.w_in.to(u.arcsec).value * scale_w
        w_out = self.w_out.to(u.arcsec).value * scale_w
        h_in = self.h_in.to(u.arcsec).value * scale_h
        h_out = self.h_out.to(u.arcsec).value * scale_h
        return RectangularAnnulus(positions=positions, w_in=w_in,
                                  w_out=w_out, h_out=h_out,
                                  h_in=h_in, theta=pixel_angle)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for parameter validation.
"""

import numpy as np
from astropy.stats import SigmaClip


class SigmaClipSentinelDefault:
    """
    A sentinel object to indicate the default value for sigma_clip.

    Parameters
    ----------
    sigma : float, optional
        The number of standard deviations for the clipping limit.

    maxiters : int, optional
        The maximum number of sigma-clipping iterations.
    """

    def __init__(self, *, sigma=3.0, maxiters=10):
        self.sigma = sigma
        self.maxiters = maxiters

    def __repr__(self):
        return (f'<default: SigmaClip(sigma={self.sigma}, '
                f'maxiters={self.maxiters})>')


def create_default_sigmaclip(*, sigma=3.0, maxiters=10):
    """
    Return a new, default SigmaClip instance.

    Parameters
    ----------
    sigma : float, optional
        The number of standard deviations for the clipping limit.

    maxiters : int, optional
        The maximum number of sigma-clipping iterations.

    Returns
    -------
    result : `~astropy.stats.SigmaClip`
        A new `~astropy.stats.SigmaClip` instance.
    """
    return SigmaClip(sigma=sigma, maxiters=maxiters)


def as_pair(name, value, *, lower_bound=None, upper_bound=None,
            check_odd=False):
    """
    Define a pair of integer values as a 1D array.

    Parameters
    ----------
    name : str
        The name of the parameter, which is used in error messages.

    value : int or int array_like
        The input value.

    lower_bound : tuple of 2 int, optional
        A tuple defining the allowed lower bound of the value. The first
        element is the bound; the second is 0 for exclusive or 1 for
        inclusive (e.g. (0, 1) means value must be >= 0).

    upper_bound : tuple of 2 int, optional
        A tuple defining the allowed upper bounds of the value along
        each axis. For each axis, if ``value`` is larger than the bound,
        it is reset to the bound. ``upper_bound`` is typically set to an
        image shape.

    check_odd : bool, optional
        Whether to raise a `ValueError` if the values are not odd along
        both axes.

    Returns
    -------
    result : (2,) `~numpy.ndarray`
        The pair as a 1D array of two integers.

    Examples
    --------
    >>> from photutils.utils._parameters import as_pair

    >>> as_pair('myparam', 4)
    array([4, 4])

    >>> as_pair('myparam', (3, 4))
    array([3, 4])

    >>> as_pair('myparam', 0, lower_bound=(0, 1))
    array([0, 0])
    """
    value = np.atleast_1d(value)

    if value.ndim != 1:
        msg = f'{name} must be 1D'
        raise ValueError(msg)

    if np.any(~np.isfinite(value)):
        msg = f'{name} must be a finite value'
        raise ValueError(msg)

    if len(value) not in (1, 2):
        msg = f'{name} must have 1 or 2 elements'
        raise ValueError(msg)
    if len(value) == 1:
        value = np.array((value[0], value[0]))

    if value.dtype.kind != 'i':
        msg = f'{name} must have integer values'
        raise ValueError(msg)
    if check_odd and np.any(value % 2 != 1):
        msg = f'{name} must have an odd value for both axes'
        raise ValueError(msg)

    if lower_bound is not None:
        if len(lower_bound) != 2:
            msg = 'lower_bound must contain only 2 elements'
            raise ValueError(msg)
        bound, inclusive = lower_bound
        if inclusive:
            oper = '>='
            mask = value < bound
        else:
            oper = '>'
            mask = value <= bound
        if np.any(mask):
            msg = f'{name} must be {oper} {bound}'
            raise ValueError(msg)

    if upper_bound is not None:
        if len(upper_bound) != 2:
            msg = 'upper_bound must contain only 2 elements'
            raise ValueError(msg)
        # If value is larger than upper_bound, set to upper_bound;
        # upper_bound is typically set to an image shape
        value = np.array((min(value[0], upper_bound[0]),
                          min(value[1], upper_bound[1])))

    return value



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Subpackage containing tools for detecting point-like (stellar) sources
or local peaks in an astronomical image.
"""

from .core import *  # noqa: F401, F403
from .daofinder import *  # noqa: F401, F403
from .irafstarfinder import *  # noqa: F401, F403
from .peakfinder import *  # noqa: F401, F403
from .starfinder import *  # noqa: F401, F403



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the _moments module.
"""

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from photutils.utils._moments import _image_moments


def test_moments():
    """
    Test _image_moments with a simple 2x2 array (raw moments).
    """
    data = np.array([[0, 1], [0, 1]])
    moments = _image_moments(data, order=2)
    result = np.array([[2, 2, 2], [1, 1, 1], [1, 1, 1]])

    assert_equal(moments, result)
    assert_allclose(moments[0, 1] / moments[0, 0], 1.0)
    assert_allclose(moments[1, 0] / moments[0, 0], 0.5)


def test_moments_central():
    """
    Test _image_moments with center=None (central moments).
    """
    data = np.array([[0, 1], [0, 1]])
    moments = _image_moments(data, center=None, order=2)
    result = np.array([[2.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.5, 0.0, 0.0]])
    assert_allclose(moments, result)


def test_moments_central_nonsquare():
    """
    Test _image_moments with center=None and a non-square array.
    """
    data = np.array([[0, 1], [0, 1], [0, 1]])
    moments = _image_moments(data, center=None, order=2)
    result = np.array([[3.0, 0.0, 0.0], [0.0, 0.0, 0.0], [2.0, 0.0, 0.0]])
    assert_allclose(moments, result)


def test_moments_central_invalid_dim():
    """
    Test that _image_moments with non-2D data raises ValueError.
    """
    data = np.arange(27).reshape(3, 3, 3)
    match = 'data must be a 2D array'
    with pytest.raises(ValueError, match=match):
        _image_moments(data, order=3)


def test_moments_central_negative_order():
    """
    Test that _image_moments with negative order raises ValueError.
    """
    data = np.array([[0, 1], [0, 1]])
    match = 'order must be non-negative'
    with pytest.raises(ValueError, match=match):
        _image_moments(data, order=-1)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the depths module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.convolution import convolve
from astropy.tests.helper import assert_quantity_allclose
from astropy.utils.exceptions import (AstropyDeprecationWarning,
                                      AstropyUserWarning)
from numpy.testing import assert_allclose

from photutils.datasets import make_100gaussians_image
from photutils.segmentation import SourceFinder, make_2dgaussian_kernel
from photutils.utils._optional_deps import HAS_SKIMAGE
from photutils.utils.depths import ImageDepth

bool_vals = (True, False)


@pytest.mark.skipif(not HAS_SKIMAGE, reason='skimage is required')
class TestImageDepth:
    def setup_class(self):
        bkg = 5.0
        data = make_100gaussians_image() - bkg
        kernel = make_2dgaussian_kernel(3.0, size=5)
        convolved_data = convolve(data, kernel)

        n_pixels = 10
        threshold = 3.2
        finder = SourceFinder(n_pixels=n_pixels, progress_bar=False)
        segment_map = finder(convolved_data, threshold)
        self.data = data
        self.mask = segment_map.make_source_mask()

    @pytest.mark.parametrize('units', bool_vals)
    @pytest.mark.parametrize('overlap', bool_vals)
    def test_image_depth(self, units, overlap):
        """
        Test ImageDepth with various unit and overlap settings.
        """
        radius = 4
        depth = ImageDepth(radius, n_sigma=5.0, n_apertures=100, n_iters=2,
                           mask_pad=5, overlap=overlap, seed=123,
                           zeropoint=23.9, progress_bar=False)

        if overlap:
            exp_limits = (72.65695364143787, 19.246807037943814)
        else:
            exp_limits = (71.07332848526178, 19.27073336332396)

        data = self.data
        fluxlim = exp_limits[0]
        if units:
            data = self.data * u.Jy
            fluxlim *= u.Jy

        limits = depth(data, self.mask)
        assert_allclose(limits[1], exp_limits[1])

        if not units:
            assert_allclose(limits[0], fluxlim)
        else:
            assert_quantity_allclose(limits[0], fluxlim)

    def test_mask_none(self):
        """
        Test ImageDepth with mask=None.
        """
        radius = 4
        depth = ImageDepth(radius, n_sigma=5.0, n_apertures=100, n_iters=2,
                           mask_pad=5, overlap=True, seed=123, zeropoint=23.9,
                           progress_bar=False)
        limits = depth(self.data, mask=None)
        assert_allclose(limits, (79.348118, 19.151158))

    def test_many_apertures(self):
        """
        Test ImageDepth with too many apertures.
        """
        radius = 4
        depth = ImageDepth(radius, n_sigma=5.0, n_apertures=5000, n_iters=2,
                           mask_pad=5, overlap=True, seed=123, zeropoint=23.9,
                           progress_bar=False)
        mask = np.zeros(self.data.shape)
        mask[:, 20:] = True
        match = 'Too many apertures for given unmasked area'
        with pytest.raises(ValueError, match=match):
            depth(self.data, mask)

        depth = ImageDepth(radius, n_sigma=5.0, n_apertures=250, n_iters=2,
                           mask_pad=5, overlap=False, seed=123, zeropoint=23.9,
                           progress_bar=False)
        mask = np.zeros(self.data.shape)
        mask[:, 100:] = True
        match = r'Unable to generate .* non-overlapping apertures'
        with pytest.warns(AstropyUserWarning, match=match):
            depth(self.data, mask)

        # Test for zero non-overlapping apertures before slow loop
        radius = 5
        depth = ImageDepth(radius, n_sigma=5.0, n_apertures=100, n_iters=2,
                           overlap=False, seed=123, zeropoint=23.9,
                           progress_bar=False)
        mask = np.zeros(self.data.shape)
        mask[:, 40:] = True
        match = r'Unable to generate .* non-overlapping apertures'
        with pytest.warns(AstropyUserWarning, match=match):
            depth(self.data, mask)

    def test_zero_data(self):
        """
        Test ImageDepth with all-zero data.
        """
        radius = 4
        depth = ImageDepth(radius, n_apertures=500, n_iters=2,
                           overlap=True, seed=123, progress_bar=False)
        data = np.zeros((300, 400))
        mask = None
        match = 'One or more flux_limit values was zero'
        with pytest.warns(AstropyUserWarning, match=match):
            limits = depth(data, mask)
        assert_allclose(limits, (0.0, np.inf))

    def test_all_masked(self):
        """
        Test ImageDepth when all pixels are masked.
        """
        radius = 4
        depth = ImageDepth(radius, n_apertures=500, n_iters=1, mask_pad=5,
                           overlap=True, seed=123, progress_bar=False)
        data = np.zeros(self.data.shape)
        mask = np.zeros(data.shape, dtype=bool)
        mask[:, 10:] = True
        match = 'There are no unmasked pixel values'
        with pytest.raises(ValueError, match=match):
            depth(data, mask)

    def test_mask_not_modified(self):
        """
        Test that the input mask is not modified in place.
        """
        radius = 4
        depth = ImageDepth(radius, n_sigma=5.0, n_apertures=100, n_iters=2,
                           mask_pad=5, overlap=True, seed=123,
                           zeropoint=23.9, progress_bar=False)
        mask_orig = np.zeros(self.data.shape, dtype=bool)
        mask_copy = mask_orig.copy()
        depth(self.data, mask_orig)
        np.testing.assert_array_equal(mask_orig, mask_copy)

        # Also when mask has no True pixels (border-only masking path)
        mask_empty = np.zeros(self.data.shape, dtype=bool)
        mask_empty_copy = mask_empty.copy()
        depth(self.data, mask_empty)
        np.testing.assert_array_equal(mask_empty, mask_empty_copy)

    def test_inputs(self):
        """
        Test ImageDepth with invalid input parameters.
        """
        match = 'aper_radius must be > 0'
        with pytest.raises(ValueError, match=match):
            ImageDepth(0.0, n_sigma=5.0, n_apertures=500, n_iters=2,
                       overlap=True, seed=123, zeropoint=23.9,
                       progress_bar=False)

        match = 'aper_radius must be > 0'
        with pytest.raises(ValueError, match=match):
            ImageDepth(-12.4, n_sigma=5.0, n_apertures=500, n_iters=2,
                       overlap=True, seed=123, zeropoint=23.9,
                       progress_bar=False)

        match = 'mask_pad must be >= 0'
        with pytest.raises(ValueError, match=match):
            ImageDepth(12.4, n_sigma=5.0, n_apertures=500, n_iters=2,
                       mask_pad=-7.1, overlap=True, seed=123, zeropoint=23.9,
                       progress_bar=False)

        match = 'sigma_clip must be a callable'
        with pytest.raises(TypeError, match=match):
            ImageDepth(4.0, n_sigma=5.0, n_apertures=500, n_iters=2,
                       sigma_clip='not_callable', progress_bar=False)

    def test_repr(self):
        """
        Test ImageDepth __repr__ output.
        """
        depth = ImageDepth(aper_radius=4, n_sigma=5.0, n_apertures=100,
                           n_iters=2, overlap=False, seed=123,
                           zeropoint=23.9, progress_bar=False)
        cls_repr = repr(depth)
        assert cls_repr.startswith(f'{depth.__class__.__name__}')

    def test_progress_bar(self):
        """
        Test running ImageDepth with progress_bar=True.
        """
        radius = 4
        depth = ImageDepth(radius, n_sigma=5.0, n_apertures=100, n_iters=1,
                           mask_pad=5, overlap=True, seed=123,
                           zeropoint=23.9, progress_bar=True)
        limits = depth(self.data, self.mask)
        assert np.isfinite(limits[0])
        assert np.isfinite(limits[1])

    def test_deprecation(self):
        """
        Test ImageDepth deprecation warnings.
        """
        depth = ImageDepth(aper_radius=4, n_sigma=5.0, n_apertures=100,
                           n_iters=2, mask_pad=5, overlap=True, seed=123,
                           zeropoint=23.9, progress_bar=False)
        match = 'attribute was deprecated'
        with pytest.warns(AstropyDeprecationWarning, match=match):
            _ = depth.napers



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tools for estimating the background and background RMS in an array of
any dimension.
"""

import abc
import warnings

import numpy as np
from astropy.stats import SigmaClip, biweight_location, biweight_scale, mad_std

from photutils.utils._deprecation import deprecated_positional_kwargs
from photutils.utils._parameters import (SigmaClipSentinelDefault,
                                         create_default_sigmaclip)
from photutils.utils._repr import make_repr
from photutils.utils._stats import nanmean, nanmedian, nanstd

__all__ = [
    'BackgroundBase',
    'BackgroundRMSBase',
    'BiweightLocationBackground',
    'BiweightScaleBackgroundRMS',
    'MADStdBackgroundRMS',
    'MMMBackground',
    'MeanBackground',
    'MedianBackground',
    'ModeEstimatorBackground',
    'SExtractorBackground',
    'StdBackgroundRMS',
]


SIGMA_CLIP = SigmaClipSentinelDefault(sigma=3.0, maxiters=10)

_SIGMA_CLIP_PARAM_DOC = (
    'sigma_clip : `astropy.stats.SigmaClip` or `None`, optional\n'
    '    A `~astropy.stats.SigmaClip` object that defines the sigma\n'
    '    clipping parameters. If `None` then no sigma clipping will be\n'
    '    performed.'
)


def _insert_sigma_clip_doc(cls):
    """
    Class decorator that replaces the ``<sigma_clip_param>`` placeholder
    in a class docstring with the shared ``sigma_clip`` parameter
    description.
    """
    cls.__doc__ = cls.__doc__.replace(
        '<sigma_clip_param>', _SIGMA_CLIP_PARAM_DOC,
    )
    return cls


def _validate_sigma_clip(sigma_clip):
    """
    Validate and generate the ``sigma_clip`` parameter.

    If ``sigma_clip`` is the sentinel default, a fresh
    `~astropy.stats.SigmaClip` instance is created from its stored
    parameters. `None` is accepted (meaning no sigma clipping). Any
    other value must be a `~astropy.stats.SigmaClip` instance.

    Parameters
    ----------
    sigma_clip : `~photutils.utils._parameters.SigmaClipSentinelDefault`,\
            `~astropy.stats.SigmaClip`, or `None`
        The value supplied to a base-class ``__init__``.

    Returns
    -------
    sigma_clip : `~astropy.stats.SigmaClip` or `None`
        A concrete `~astropy.stats.SigmaClip` instance, or `None`.
    """
    if sigma_clip is SIGMA_CLIP:
        return create_default_sigmaclip(sigma=SIGMA_CLIP.sigma,
                                        maxiters=SIGMA_CLIP.maxiters)
    if not isinstance(sigma_clip, SigmaClip) and sigma_clip is not None:
        msg = 'sigma_clip must be an astropy SigmaClip instance or None'
        raise TypeError(msg)

    return sigma_clip


def _prepare_data(sigma_clip, data, axis):
    """
    Prepare input data for a background estimation step.

    Applies sigma clipping when a `~astropy.stats.SigmaClip` instance is
    provided, or fills masked-array fill values with NaN when the input
    is a `~numpy.ma.MaskedArray` and sigma clipping is disabled.

    Parameters
    ----------
    sigma_clip : `~astropy.stats.SigmaClip` or `None`
        The sigma-clipping object to apply. If `None`, no clipping is
        performed.

    data : array_like or `~numpy.ma.MaskedArray`
        The input data array.

    axis : int, tuple of int, or `None`
        The axis along which sigma clipping is applied.

    Returns
    -------
    data : `~numpy.ndarray`
        The prepared data array, with masked or clipped values replaced
        by NaN.
    """
    if sigma_clip is not None:
        return sigma_clip(data, axis=axis, masked=False)

    if isinstance(data, np.ma.MaskedArray):
        # convert to ndarray with masked values replaced by NaN
        return data.filled(np.nan)

    return data


def _apply_masked(result, masked):
    """
    Optionally wrap NaN values in a masked array.

    Parameters
    ----------
    result : `~numpy.ndarray` or scalar
        The computed background or background RMS value(s).

    masked : bool
        If `True` and ``result`` is an `~numpy.ndarray`, return a
        `~numpy.ma.MaskedArray` with NaN values masked. Otherwise return
        ``result`` unchanged.

    Returns
    -------
    result : `~numpy.ndarray`, `~numpy.ma.MaskedArray`, or scalar
        The result, optionally wrapped as a masked array.
    """
    if masked and isinstance(result, np.ndarray):
        return np.ma.masked_where(np.isnan(result), result)
    return result


class _BackgroundCommonBase:
    """
    Internal mixin providing shared infrastructure for `BackgroundBase`
    and `BackgroundRMSBase`.

    This class is not part of the public API and should not be
    instantiated directly or subclassed outside of this module.
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, sigma_clip=SIGMA_CLIP):
        self.sigma_clip = _validate_sigma_clip(sigma_clip)

    def __repr__(self):
        return make_repr(self, ('sigma_clip',))


class BackgroundBase(_BackgroundCommonBase, abc.ABC):
    """
    Base class for classes that estimate scalar background values.
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __call__(self, data, axis=None, masked=False):
        return self.calc_background(data, axis=axis, masked=masked)

    @abc.abstractmethod
    def calc_background(self, data, *, axis=None, masked=False):
        """
        Calculate the background value.

        Parameters
        ----------
        data : array_like or `~numpy.ma.MaskedArray`
            The array for which to calculate the background value.

        axis : int or `None`, optional
            The array axis along which the background is calculated. If
            `None`, then the entire array is used.

        masked : bool, optional
            If `True`, then a `~numpy.ma.MaskedArray` is returned. If
            `False`, then a `~numpy.ndarray` is returned, where masked
            values have a value of NaN. The default is `False`.

        Returns
        -------
        result : float, `~numpy.ndarray`, or `~numpy.ma.MaskedArray`
            The calculated background value. If ``masked`` is
            `False`, then a `~numpy.ndarray` is returned, otherwise a
            `~numpy.ma.MaskedArray` is returned. A scalar result is
            always returned as a float.
        """


class BackgroundRMSBase(_BackgroundCommonBase, abc.ABC):
    """
    Base class for classes that estimate scalar background RMS values.
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __call__(self, data, axis=None, masked=False):
        return self.calc_background_rms(data, axis=axis, masked=masked)

    @abc.abstractmethod
    def calc_background_rms(self, data, *, axis=None, masked=False):
        """
        Calculate the background RMS value.

        Parameters
        ----------
        data : array_like or `~numpy.ma.MaskedArray`
            The array for which to calculate the background RMS value.

        axis : int or `None`, optional
            The array axis along which the background RMS is calculated.
            If `None`, then the entire array is used.

        masked : bool, optional
            If `True`, then a `~numpy.ma.MaskedArray` is returned. If
            `False`, then a `~numpy.ndarray` is returned, where masked
            values have a value of NaN. The default is `False`.

        Returns
        -------
        result : float, `~numpy.ndarray`, or `~numpy.ma.MaskedArray`
            The calculated background RMS value. If ``masked`` is
            `False`, then a `~numpy.ndarray` is returned, otherwise a
            `~numpy.ma.MaskedArray` is returned. A scalar result is
            always returned as a float.
        """


@_insert_sigma_clip_doc
class MeanBackground(BackgroundBase):
    """
    Class to calculate the background in an array as the (sigma-clipped)
    mean.

    Parameters
    ----------
    <sigma_clip_param>

    Examples
    --------
    >>> from astropy.stats import SigmaClip
    >>> from photutils.background import MeanBackground
    >>> data = np.arange(100)
    >>> sigma_clip = SigmaClip(sigma=3.0)
    >>> bkg = MeanBackground(sigma_clip=sigma_clip)

    The background value can be calculated by using the
    `calc_background` method, e.g.:

    >>> bkg_value = bkg.calc_background(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5

    Alternatively, the background value can be calculated by calling the
    class instance as a function, e.g.:

    >>> bkg_value = bkg(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def calc_background(self, data, axis=None, masked=False):
        data = _prepare_data(self.sigma_clip, data, axis)
        # Ignore RuntimeWarning where axis is all NaN
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            result = nanmean(data, axis=axis)
        return _apply_masked(result, masked)


@_insert_sigma_clip_doc
class MedianBackground(BackgroundBase):
    """
    Class to calculate the background in an array as the (sigma-clipped)
    median.

    Parameters
    ----------
    <sigma_clip_param>

    Examples
    --------
    >>> from astropy.stats import SigmaClip
    >>> from photutils.background import MedianBackground
    >>> data = np.arange(100)
    >>> sigma_clip = SigmaClip(sigma=3.0)
    >>> bkg = MedianBackground(sigma_clip=sigma_clip)

    The background value can be calculated by using the
    `calc_background` method, e.g.:

    >>> bkg_value = bkg.calc_background(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5

    Alternatively, the background value can be calculated by calling the
    class instance as a function, e.g.:

    >>> bkg_value = bkg(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def calc_background(self, data, axis=None, masked=False):
        data = _prepare_data(self.sigma_clip, data, axis)
        # Ignore RuntimeWarning where axis is all NaN
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            result = nanmedian(data, axis=axis)
        return _apply_masked(result, masked)


@_insert_sigma_clip_doc
class ModeEstimatorBackground(BackgroundBase):
    """
    Class to calculate the background in an array using a mode estimator
    of the form ``(median_factor * median) - (mean_factor * mean)``.

    Parameters
    ----------
    median_factor : float, optional
        The multiplicative factor for the median value. Defaults to 3.

    mean_factor : float, optional
        The multiplicative factor for the mean value. Defaults to 2.

    <sigma_clip_param>

    Examples
    --------
    >>> from astropy.stats import SigmaClip
    >>> from photutils.background import ModeEstimatorBackground
    >>> data = np.arange(100)
    >>> sigma_clip = SigmaClip(sigma=3.0)
    >>> bkg = ModeEstimatorBackground(median_factor=3.0, mean_factor=2.0,
    ...                               sigma_clip=sigma_clip)

    The background value can be calculated by using the
    `calc_background` method, e.g.:

    >>> bkg_value = bkg.calc_background(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5

    Alternatively, the background value can be calculated by calling the
    class instance as a function, e.g.:

    >>> bkg_value = bkg(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, median_factor=3.0, mean_factor=2.0,
                 sigma_clip=SIGMA_CLIP):
        super().__init__(sigma_clip=sigma_clip)
        self.median_factor = median_factor
        self.mean_factor = mean_factor

    def __repr__(self):
        params = ('median_factor', 'mean_factor', 'sigma_clip')
        return make_repr(self, params)

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def calc_background(self, data, axis=None, masked=False):
        data = _prepare_data(self.sigma_clip, data, axis)
        # Ignore RuntimeWarning where axis is all NaN
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            result = ((self.median_factor * nanmedian(data, axis=axis))
                      - (self.mean_factor * nanmean(data, axis=axis)))
        return _apply_masked(result, masked)


@_insert_sigma_clip_doc
class MMMBackground(ModeEstimatorBackground):
    """
    Class to calculate the background in an array using the DAOPHOT MMM
    algorithm.

    The background is calculated using a mode estimator of the form
    ``(3 * median) - (2 * mean)``.

    Parameters
    ----------
    <sigma_clip_param>

    Examples
    --------
    >>> from astropy.stats import SigmaClip
    >>> from photutils.background import MMMBackground
    >>> data = np.arange(100)
    >>> sigma_clip = SigmaClip(sigma=3.0)
    >>> bkg = MMMBackground(sigma_clip=sigma_clip)

    The background value can be calculated by using the
    `calc_background` method, e.g.:

    >>> bkg_value = bkg.calc_background(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5

    Alternatively, the background value can be calculated by calling the
    class instance as a function, e.g.:

    >>> bkg_value = bkg(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, sigma_clip=SIGMA_CLIP):
        super().__init__(median_factor=3.0, mean_factor=2.0,
                         sigma_clip=sigma_clip)


@_insert_sigma_clip_doc
class SExtractorBackground(BackgroundBase):
    """
    Class to calculate the background in an array using the Source
    Extractor algorithm.

    The background is calculated using a mode estimator of the form
    ``(2.5 * median) - (1.5 * mean)``. If ``(mean - median) / std >
    0.3`` then the median is used instead.

    .. _SourceExtractor: https://sextractor.readthedocs.io/en/latest/

    Parameters
    ----------
    <sigma_clip_param>

    Examples
    --------
    >>> from astropy.stats import SigmaClip
    >>> from photutils.background import SExtractorBackground
    >>> data = np.arange(100)
    >>> sigma_clip = SigmaClip(sigma=3.0)
    >>> bkg = SExtractorBackground(sigma_clip=sigma_clip)

    The background value can be calculated by using the
    `calc_background` method, e.g.:

    >>> bkg_value = bkg.calc_background(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5

    Alternatively, the background value can be calculated by calling the
    class instance as a function, e.g.:

    >>> bkg_value = bkg(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def calc_background(self, data, axis=None, masked=False):
        data = _prepare_data(self.sigma_clip, data, axis)
        # Ignore RuntimeWarning where axis is all NaN
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)

            _median = np.atleast_1d(nanmedian(data, axis=axis))
            _mean = np.atleast_1d(nanmean(data, axis=axis))
            _std = np.atleast_1d(nanstd(data, axis=axis))
            result = (2.5 * _median) - (1.5 * _mean)

            # Set the background to the mean where the std is zero
            mean_mask = _std == 0
            result[mean_mask] = _mean[mean_mask]

            # Set the background to the median when the absolute
            # difference between the mean and median divided by the
            # standard deviation is greater than or equal to 0.3
            med_mask = (np.abs(_mean - _median) / _std) >= 0.3
            mask = np.logical_and(med_mask, np.logical_not(mean_mask))
            result[mask] = _median[mask]

            # If result is a scalar, return it as a float
            if result.shape == (1,) and axis is None:
                result = result[0]

        return _apply_masked(result, masked)


@_insert_sigma_clip_doc
class BiweightLocationBackground(BackgroundBase):
    """
    Class to calculate the background in an array using the biweight
    location.

    Parameters
    ----------
    c : float, optional
        Tuning constant for the biweight estimator. Default value is
        6.0.

    M : float, optional
        Initial guess for the biweight location. Default value is
        `None`.

    <sigma_clip_param>

    Examples
    --------
    >>> from astropy.stats import SigmaClip
    >>> from photutils.background import BiweightLocationBackground
    >>> data = np.arange(100)
    >>> sigma_clip = SigmaClip(sigma=3.0)
    >>> bkg = BiweightLocationBackground(sigma_clip=sigma_clip)

    The background value can be calculated by using the
    `calc_background` method, e.g.:

    >>> bkg_value = bkg.calc_background(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5

    Alternatively, the background value can be calculated by calling the
    class instance as a function, e.g.:

    >>> bkg_value = bkg(data)
    >>> print(bkg_value)  # doctest: +FLOAT_CMP
    49.5
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, c=6.0, M=None, sigma_clip=SIGMA_CLIP):
        super().__init__(sigma_clip=sigma_clip)
        self.c = c
        self.M = M

    def __repr__(self):
        params = ('c', 'M', 'sigma_clip')
        return make_repr(self, params)

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def calc_background(self, data, axis=None, masked=False):
        data = _prepare_data(self.sigma_clip, data, axis)
        # Ignore RuntimeWarning where axis is all NaN
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            result = biweight_location(data, c=self.c, M=self.M, axis=axis,
                                       ignore_nan=True)
        return _apply_masked(result, masked)


@_insert_sigma_clip_doc
class StdBackgroundRMS(BackgroundRMSBase):
    """
    Class to calculate the background RMS in an array as the (sigma-
    clipped) standard deviation.

    Parameters
    ----------
    <sigma_clip_param>

    Examples
    --------
    >>> from astropy.stats import SigmaClip
    >>> from photutils.background import StdBackgroundRMS
    >>> data = np.arange(100)
    >>> sigma_clip = SigmaClip(sigma=3.0)
    >>> bkgrms = StdBackgroundRMS(sigma_clip=sigma_clip)

    The background RMS value can be calculated by using the
    `calc_background_rms` method, e.g.:

    >>> bkgrms_value = bkgrms.calc_background_rms(data)
    >>> print(bkgrms_value)  # doctest: +FLOAT_CMP
    28.86607004772212

    Alternatively, the background RMS value can be calculated by calling
    the class instance as a function, e.g.:

    >>> bkgrms_value = bkgrms(data)
    >>> print(bkgrms_value)  # doctest: +FLOAT_CMP
    28.86607004772212
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def calc_background_rms(self, data, axis=None, masked=False):
        data = _prepare_data(self.sigma_clip, data, axis)
        # Ignore RuntimeWarning where axis is all NaN
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            result = nanstd(data, axis=axis)
        return _apply_masked(result, masked)


@_insert_sigma_clip_doc
class MADStdBackgroundRMS(BackgroundRMSBase):
    r"""
    Class to calculate the background RMS in an array as using the
    `median absolute deviation (MAD)
    <https://en.wikipedia.org/wiki/Median_absolute_deviation>`_.

    The standard deviation estimator is given by:

    .. math::

        \sigma \approx \frac{{\textrm{{MAD}}}}{{\Phi^{{-1}}(3/4)}}
            \approx 1.4826 \ \textrm{{MAD}}

    where :math:`\Phi^{{-1}}(P)` is the normal inverse cumulative
    distribution function evaluated at probability :math:`P = 3/4`.

    Parameters
    ----------
    <sigma_clip_param>

    Examples
    --------
    >>> from astropy.stats import SigmaClip
    >>> from photutils.background import MADStdBackgroundRMS
    >>> data = np.arange(100)
    >>> sigma_clip = SigmaClip(sigma=3.0)
    >>> bkgrms = MADStdBackgroundRMS(sigma_clip=sigma_clip)

    The background RMS value can be calculated by using the
    `calc_background_rms` method, e.g.:

    >>> bkgrms_value = bkgrms.calc_background_rms(data)
    >>> print(bkgrms_value)  # doctest: +FLOAT_CMP
    37.06505546264005

    Alternatively, the background RMS value can be calculated by calling
    the class instance as a function, e.g.:

    >>> bkgrms_value = bkgrms(data)
    >>> print(bkgrms_value)  # doctest: +FLOAT_CMP
    37.06505546264005
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def calc_background_rms(self, data, axis=None, masked=False):
        data = _prepare_data(self.sigma_clip, data, axis)
        # Ignore RuntimeWarning where axis is all NaN
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            result = mad_std(data, axis=axis, ignore_nan=True)
        return _apply_masked(result, masked)


@_insert_sigma_clip_doc
class BiweightScaleBackgroundRMS(BackgroundRMSBase):
    """
    Class to calculate the background RMS in an array as the (sigma-
    clipped) biweight scale.

    Parameters
    ----------
    c : float, optional
        Tuning constant for the biweight estimator. Default value is
        9.0.

    M : float, optional
        Initial guess for the biweight location. Default value is
        `None`.

    <sigma_clip_param>

    Examples
    --------
    >>> from astropy.stats import SigmaClip
    >>> from photutils.background import BiweightScaleBackgroundRMS
    >>> data = np.arange(100)
    >>> sigma_clip = SigmaClip(sigma=3.0)
    >>> bkgrms = BiweightScaleBackgroundRMS(sigma_clip=sigma_clip)

    The background RMS value can be calculated by using the
    `calc_background_rms` method, e.g.:

    >>> bkgrms_value = bkgrms.calc_background_rms(data)
    >>> print(bkgrms_value)  # doctest: +FLOAT_CMP
    30.09433848589339

    Alternatively, the background RMS value can be calculated by calling
    the class instance as a function, e.g.:

    >>> bkgrms_value = bkgrms(data)
    >>> print(bkgrms_value)  # doctest: +FLOAT_CMP
    30.09433848589339
    """

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def __init__(self, c=9.0, M=None, sigma_clip=SIGMA_CLIP):
        super().__init__(sigma_clip=sigma_clip)
        self.c = c
        self.M = M

    def __repr__(self):
        params = ('c', 'M', 'sigma_clip')
        return make_repr(self, params)

    @deprecated_positional_kwargs(since='3.0', until='4.0')
    def calc_background_rms(self, data, axis=None, masked=False):
        data = _prepare_data(self.sigma_clip, data, axis)
        # Ignore RuntimeWarning where axis is all NaN
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', RuntimeWarning)
            result = biweight_scale(data, c=self.c, M=self.M, axis=axis,
                                    ignore_nan=True)
        return _apply_masked(result, masked)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the _optional_deps module.
"""

import importlib
from unittest.mock import patch

import pytest

import photutils.utils._optional_deps as od_mod
from photutils.utils._optional_deps import (_deps_by_key, _dist_to_has_key,
                                            _get_optional_deps, _pkg_dist_name)


def _clear_cache(*names):
    """
    Remove named entries from the module-level cache.

    This is necessary to ensure that tests of the caching behavior are
    valid and not affected by previous tests.
    """
    for name in names:
        od_mod._cache.pop(name, None)


class TestAll:
    def test_all_is_list(self):
        """
        Test that ``__all__`` is a list.
        """
        assert isinstance(od_mod.__all__, list)

    def test_all_entries_start_with_has(self):
        """
        Test that every entry in ``__all__`` starts with ``HAS_``.
        """
        for name in od_mod.__all__:
            assert name.startswith('HAS_')

    def test_all_is_sorted(self):
        """
        Test that ``__all__`` is sorted alphabetically.
        """
        assert od_mod.__all__ == sorted(od_mod.__all__)

    def test_skimage_always_in_all(self):
        """
        Test that ``HAS_SKIMAGE`` is always present in ``__all__``.
        """
        assert 'HAS_SKIMAGE' in od_mod.__all__

    def test_all_matches_deps_by_key(self):
        """
        Test that ``__all__`` matches the keys in ``_deps_by_key``.
        """
        expected = sorted(f'HAS_{k}' for k in _deps_by_key)
        assert od_mod.__all__ == expected


class TestHasAttributes:
    def test_installed_package_returns_true(self):
        """
        Test that a ``HAS_*`` attribute is `True` when the import
        succeeds.
        """
        _clear_cache('HAS_MATPLOTLIB')
        try:
            with patch.object(importlib, 'import_module'):
                assert od_mod.HAS_MATPLOTLIB is True
        finally:
            _clear_cache('HAS_MATPLOTLIB')

    def test_import_error_returns_false(self):
        """
        Test that a ``HAS_*`` attribute is `False` when the import
        raises `ImportError`.
        """
        _clear_cache('HAS_MATPLOTLIB')
        try:
            with patch.object(importlib, 'import_module',
                              side_effect=ImportError):
                result = od_mod.HAS_MATPLOTLIB
            assert result is False
        finally:
            _clear_cache('HAS_MATPLOTLIB')

    def test_result_is_bool(self):
        """
        Test that ``HAS_*`` attributes are `bool`.
        """
        _clear_cache('HAS_MATPLOTLIB')
        try:
            assert isinstance(od_mod.HAS_MATPLOTLIB, bool)
        finally:
            _clear_cache('HAS_MATPLOTLIB')


class TestCaching:
    def test_value_is_cached(self):
        """
        Test that the result of a ``HAS_*`` lookup is stored in the
        cache.
        """
        _clear_cache('HAS_MATPLOTLIB')
        try:
            _ = od_mod.HAS_MATPLOTLIB
            assert 'HAS_MATPLOTLIB' in od_mod._cache
        finally:
            _clear_cache('HAS_MATPLOTLIB')

    def test_cached_value_returned_without_reimport(self):
        """
        Test that a cached value is returned without calling
        ``import_module`` again.
        """
        _clear_cache('HAS_MATPLOTLIB')
        od_mod._cache['HAS_MATPLOTLIB'] = True
        try:
            with patch.object(
                importlib, 'import_module',
                side_effect=AssertionError('should not be called'),
            ):
                assert od_mod.HAS_MATPLOTLIB is True
        finally:
            _clear_cache('HAS_MATPLOTLIB')


class TestAttributeErrors:
    def test_typo_raises(self):
        """
        Test that a plausible typo like ``HAS_SKIIMAGE`` raises
        `AttributeError`.
        """
        match = 'HAS_SKIIMAGE'
        with pytest.raises(AttributeError, match=match):
            _ = od_mod.HAS_SKIIMAGE

    def test_non_dependency_raises(self):
        """
        Test that ``HAS_NUMPY`` raises `AttributeError` because numpy is
        not an optional dependency.
        """
        if 'NUMPY' not in od_mod._deps_by_key:
            match = 'HAS_NUMPY'
            with pytest.raises(AttributeError, match=match):
                _ = od_mod.HAS_NUMPY

    def test_arbitrary_name_raises(self):
        """
        Test that a completely unknown attribute raises
        `AttributeError`.
        """
        with pytest.raises(AttributeError):
            _ = od_mod.SOME_UNKNOWN_ATTRIBUTE

    def test_has_prefix_only_raises(self):
        """
        Test that ``HAS_`` with no suffix raises `AttributeError`.
        """
        with pytest.raises(AttributeError):
            _ = od_mod.HAS_


class TestDistToHasKey:
    def test_simple(self):
        """
        Test that a simple distribution name is uppercased.
        """
        assert _dist_to_has_key('matplotlib') == 'MATPLOTLIB'

    def test_hyphenated(self):
        """
        Test that a hyphenated distribution name is converted via the
        import-name lookup (e.g., ``scikit-image`` -> ``SKIMAGE``).
        """
        assert _dist_to_has_key('scikit-image') == 'SKIMAGE'

    def test_dotted(self):
        """
        Test that dots in a distribution name are replaced with
        underscores.
        """
        assert _dist_to_has_key('stsci.stimage') == 'STSCI_STIMAGE'

    def test_mixed_case(self):
        """
        Test that mixed-case distribution names are uppercased.
        """
        assert _dist_to_has_key('MyPackage') == 'MYPACKAGE'


class TestGetOptionalDeps:
    def test_returns_sorted_list(self):
        """
        Test that ``_get_optional_deps`` returns a sorted list.
        """
        result = _get_optional_deps(_pkg_dist_name, extra='all')
        assert isinstance(result, list)
        assert result == sorted(result)

    def test_scikit_image_present(self):
        """
        Test that ``scikit-image`` appears in the optional
        dependencies.
        """
        result = _get_optional_deps(_pkg_dist_name, extra='all')
        assert 'scikit-image' in result

    def test_returns_dist_names_not_import_names(self):
        """
        Test that the returned names are distribution names, not import
        names (e.g., ``scikit-image`` instead of ``skimage``).
        """
        result = _get_optional_deps(_pkg_dist_name, extra='all')
        assert 'scikit-image' in result
        assert 'skimage' not in result



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Nan-ignoring statistical functions, using bottleneck for performance if
available.

When bottleneck is installed, it is used only for float64 arrays. For
other dtypes (e.g., float32), NumPy is used instead to work around known
accuracy issues in bottleneck (see bottleneck issues #379 and #462, and
astropy issues #17185 and #11492).
"""

from functools import partial

import numpy as np
from astropy.units import Quantity

from photutils.utils._optional_deps import HAS_BOTTLENECK

_STAT_NAMES = (
    'nansum', 'nanmin', 'nanmax', 'nanmean', 'nanmedian', 'nanstd', 'nanvar',
)

if HAS_BOTTLENECK:
    import bottleneck as bn

    def _move_tuple_axes_last(array, axis):
        """
        Move the specified axes of a NumPy array to the last positions
        and combine them.

        Bottleneck can only take integer axis, not tuple, so this
        function takes all the axes to be operated on and combines them
        into the last dimension of the array so that we can then use
        axis=-1.

        Parameters
        ----------
        array : `~numpy.ndarray`
            The input array.

        axis : tuple of int
            The axes on which to move and combine.

        Returns
        -------
        array_new : `~numpy.ndarray`
            Array with the axes being operated on moved into the last
            dimension.
        """
        other_axes = tuple(i for i in range(array.ndim) if i not in axis)

        # Move the specified axes to the last positions
        array_new = np.transpose(array, other_axes + axis)

        # Reshape the array by combining the moved axes
        return array_new.reshape((*array_new.shape[:len(other_axes)], -1))

    def _apply_bottleneck(function, array, axis=None, **kwargs):
        """
        Wrap a bottleneck function to handle tuple axis.

        This function also takes care to ensure the output is of the
        expected type, i.e., a quantity, numpy array, or numpy scalar.

        Parameters
        ----------
        function : callable
            The bottleneck function to apply.

        array : `~numpy.ndarray`
            The array on which to operate.

        axis : int or tuple of int, optional
            The axis or axes on which to operate.

        **kwargs : dict, optional
            Additional keyword arguments to pass to the bottleneck
            function.

        Returns
        -------
        result : `~numpy.ndarray` or float
            The result of the bottleneck function when called with the
            ``array``, ``axis``, and ``kwargs``.
        """
        if isinstance(axis, tuple):
            array = _move_tuple_axes_last(array, axis=axis)
            axis = -1

        result = function(array, axis=axis, **kwargs)
        if isinstance(array, Quantity):
            if function == bn.nanvar:
                result <<= array.unit ** 2
            else:
                result = array.__array_wrap__(result)
            return result

        if isinstance(result, float):
            # For compatibility with numpy, always return a numpy scalar.
            return np.float64(result)

        return result

    bn_funcs = {
        name: partial(_apply_bottleneck, getattr(bn, name))
        for name in _STAT_NAMES
    }
    np_funcs = {name: getattr(np, name) for name in _STAT_NAMES}

    class _DtypeDispatch:
        """
        Dispatcher that routes to bottleneck or numpy based on dtype.

        This is done to workaround known accuracy bugs in Bottleneck
        affecting float32 calculations. See the following issues for
        more details:

        * https://github.com/pydata/bottleneck/issues/379
        * https://github.com/pydata/bottleneck/issues/462
        * https://github.com/astropy/astropy/issues/17185
        * https://github.com/astropy/astropy/issues/11492
        """

        def __init__(self, func_name):
            self.func_name = func_name

        def __repr__(self):
            return f'_DtypeDispatch({self.func_name!r})'

        def __call__(self, *args, **kwargs):
            dt = args[0].dtype
            if dt.kind == 'f' and dt.itemsize == 8:
                return bn_funcs[self.func_name](*args, **kwargs)
            return np_funcs[self.func_name](*args, **kwargs)

    (nansum, nanmin, nanmax, nanmean, nanmedian, nanstd, nanvar) = (
        _DtypeDispatch(name) for name in _STAT_NAMES
    )

else:
    (nansum, nanmin, nanmax, nanmean, nanmedian, nanstd, nanvar) = (
        getattr(np, name) for name in _STAT_NAMES
    )



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
Tests for the images module.
"""

import astropy.units as u
import numpy as np
import pytest
from astropy.modeling.models import Moffat2D, Polynomial2D
from astropy.table import QTable
from numpy.testing import assert_allclose

from photutils.datasets import make_model_image
from photutils.datasets.images import _model_shape_from_bbox
from photutils.psf import (CircularGaussianPSF, CircularGaussianSigmaPRF,
                           ImagePSF)


def test_make_model_image():
    """
    Test the basic functionality of make_model_image.
    """
    params = QTable()
    params['x_0'] = [50, 70, 90]
    params['y_0'] = [50, 50, 50]
    params['gamma'] = [1.7, 2.32, 5.8]
    params['alpha'] = [2.9, 5.7, 4.6]
    model = Moffat2D(amplitude=1)
    shape = (300, 500)
    model_shape = (11, 11)
    image = make_model_image(shape, model, params, model_shape=model_shape)
    assert image.shape == shape
    assert image.sum() > 1

    # Test variable model shape
    params['model_shape'] = [9, 7, 11]
    image = make_model_image(shape, model, params, model_shape=model_shape)
    assert image.shape == shape
    assert image.sum() > 1

    # Test local_bkg
    params['local_bkg'] = [1, 2, 3]
    image = make_model_image(shape, model, params, model_shape=model_shape)
    assert image.shape == shape
    assert image.sum() > 1


def test_make_model_image_units():
    """
    Test that the model image is created with the correct units when the
    flux column has units.
    """
    unit = u.Jy
    params = QTable()
    params['x_0'] = [30, 50, 70.5]
    params['y_0'] = [50, 50, 50.5]
    params['flux'] = [1, 2, 3] * unit
    model = CircularGaussianSigmaPRF(sigma=1.5)
    shape = (300, 500)
    model_shape = (11, 11)
    image = make_model_image(shape, model, params, model_shape=model_shape)
    assert image.shape == shape
    assert isinstance(image, u.Quantity)
    assert image.unit == unit
    assert model.flux == 1.0  # Default flux (unchanged)

    params['local_bkg'] = [0.1, 0.2, 0.3] * unit
    image = make_model_image(shape, model, params, model_shape=model_shape)
    assert image.shape == shape
    assert isinstance(image, u.Quantity)
    assert image.unit == unit

    match = 'The local_bkg column must have the same flux units'
    params['local_bkg'] = [0.1, 0.2, 0.3]
    with pytest.raises(ValueError, match=match):
        make_model_image(shape, model, params, model_shape=model_shape)


def test_make_model_image_units_no_overlap():
    """
    Test that the model image is created with the correct units when
    there is no overlap between the model and the image.
    """
    unit = u.Jy
    params = QTable()
    params['x_0'] = [50, 70.5]
    params['y_0'] = [50, 50.5]
    params['flux'] = [2, 3] * unit
    model = CircularGaussianSigmaPRF(sigma=1.5)
    shape = (10, 12)
    image = make_model_image(shape, model, params)
    assert image.shape == shape
    assert isinstance(image, u.Quantity)
    assert image.unit == unit
    assert model.flux == 1.0  # Default flux (unchanged)

    params['flux'] = [2, 3]
    image = make_model_image(shape, model, params)
    assert image.shape == shape
    assert not isinstance(image, u.Quantity)
    assert model.flux == 1.0  # Default flux (unchanged)


def test_make_model_image_discretize_method():
    """
    Test the model image when using different discretization methods.
    """
    params = QTable()
    params['x_0'] = [50, 70, 90]
    params['y_0'] = [50, 50, 50]
    params['gamma'] = [1.7, 2.32, 5.8]
    params['alpha'] = [2.9, 5.7, 4.6]
    model = Moffat2D(amplitude=1)
    shape = (300, 500)
    model_shape = (11, 11)
    for method in ('interp', 'oversample'):
        image = make_model_image(shape, model, params, model_shape=model_shape,
                                 discretize_method=method)
        assert image.shape == shape
        assert image.sum() > 1


def test_make_model_image_no_overlap():
    """
    Test the model image when there is no overlap between the model and
    the image.
    """
    params = QTable()
    params['x_0'] = [50]
    params['y_0'] = [50]
    params['gamma'] = [1.7]
    params['alpha'] = [2.9]
    model = Moffat2D(amplitude=1)
    shape = (10, 10)
    model_shape = (3, 3)
    data = make_model_image(shape, model, params, model_shape=model_shape)
    assert data.shape == shape
    assert np.sum(data) == 0


def test_make_model_image_inputs():
    """
    Test that the appropriate exceptions are raised for invalid inputs.
    """
    match = 'shape must be a 2-tuple'
    with pytest.raises(ValueError, match=match):
        make_model_image(100, Moffat2D(), QTable())

    match = 'model must be a Model instance'
    with pytest.raises(TypeError, match=match):
        make_model_image((100, 100), None, QTable())

    match = 'model must be a 2D model'
    model = Moffat2D()
    model.n_inputs = 1
    with pytest.raises(ValueError, match=match):
        make_model_image((100, 100), model, QTable())

    match = 'params_table must be an astropy Table'
    model = Moffat2D()
    with pytest.raises(TypeError, match=match):
        make_model_image((100, 100), model, None)

    match = 'not in model parameter names'
    model = Moffat2D()
    with pytest.raises(ValueError, match=match):
        make_model_image((100, 100), model, QTable(), x_name='invalid')

    match = 'not in params_table column names'
    model = Moffat2D()
    with pytest.raises(ValueError, match=match):
        make_model_image((100, 100), model, QTable(), y_name='invalid')

    model = Moffat2D()
    params = QTable()
    with pytest.raises(ValueError, match=match):
        make_model_image((100, 100), model, params)

    model = Moffat2D()
    params = QTable()
    params['x_0'] = [50, 70, 90]
    with pytest.raises(ValueError, match=match):
        make_model_image((100, 100), model, params)

    match = 'model_shape must be specified if the model does not have'
    params = QTable()
    params['x_0'] = [50]
    params['y_0'] = [50]
    params['gamma'] = [1.7]
    params['alpha'] = [2.9]
    model = Moffat2D(amplitude=1)
    shape = (100, 100)
    with pytest.raises(ValueError, match=match):
        make_model_image(shape, model, params)


def test_make_model_image_bbox():
    """
    Test the model image when using a PSF model that has a bounding box
    and the bbox_factor keyword to control the size of the bounding box.
    """
    model1 = CircularGaussianPSF(x_0=50, y_0=50, fwhm=10)
    yy, xx = np.mgrid[:101, :101]
    model2 = ImagePSF(model1(xx, yy), x_0=50, y_0=50)

    params = QTable()
    params['x_0'] = [50, 70, 90]
    params['y_0'] = [50, 50, 50]
    shape = (100, 151)
    image1 = make_model_image(shape, model2, params, bbox_factor=10)
    image2 = make_model_image(shape, model2, params, bbox_factor=None)
    assert_allclose(image1, image2)

    image3 = make_model_image(shape, model1, params, bbox_factor=10)
    image4 = make_model_image(shape, model1, params, bbox_factor=None)
    assert_allclose(image3, image4)

    model1.bbox_factor = 10
    image5 = make_model_image(shape, model1, params)
    assert np.sum(image5) > np.sum(image4)
    assert_allclose(image3, image4)


def test_make_model_image_params_map():
    """
    Test the model image when using a parameter mapping to map the model
    parameter names to different column names in the input table.
    """
    params = QTable()
    params['x_0'] = [50, 70, 90]
    params['y_0'] = [50, 50, 50]
    params['gamma'] = [1.7, 2.32, 5.8]
    params['alpha'] = [2.9, 5.7, 4.6]
    model = Moffat2D(amplitude=1)
    shape = (300, 500)
    model_shape = (11, 11)
    image = make_model_image(shape, model, params, model_shape=model_shape)

    params = QTable()
    params['x_0'] = [50, 70, 90]
    params['y_0'] = [50, 50, 50]
    params['gamma2'] = [1.7, 2.32, 5.8]
    params['alpha4'] = [2.9, 5.7, 4.6]
    params_map = {'gamma': 'gamma2', 'alpha': 'alpha4'}
    model = Moffat2D(amplitude=1)
    shape = (300, 500)
    model_shape = (11, 11)
    image2 = make_model_image(shape, model, params, model_shape=model_shape,
                              params_map=params_map)
    assert_allclose(image, image2)


def test_make_model_image_nonfinite():
    """
    Test the model image when the input table contains non-finite
    values.
    """
    params = QTable()
    params['x_0'] = [50, np.nan, 90, 100]
    params['y_0'] = [50, 50, 50, 50]
    params['gamma'] = [1.7, 2.32, 5.8, np.inf]
    params['alpha'] = [2.9, 5.7, 4.6, 3.1]
    model = Moffat2D(amplitude=1)
    shape = (300, 500)
    model_shape = (11, 11)
    image = make_model_image(shape, model, params, model_shape=model_shape)
    assert image.shape == shape
    assert image.sum() < 33
    assert image[50, 100] == 0

    # All invalid sources
    params = QTable()
    params['x_0'] = [50, np.nan, 90, 100]
    params['y_0'] = [-np.inf, 50, 50, 50]
    params['gamma'] = [1.7, 2.32, 5.8, np.inf]
    params['alpha'] = [2.9, 5.7, np.nan, 3.1]
    model = Moffat2D(amplitude=1)
    shape = (300, 500)
    model_shape = (11, 11)
    image = make_model_image(shape, model, params, model_shape=model_shape)
    assert image.shape == shape
    assert image.sum() == 0


def test_make_model_image_progress_bar():
    """
    Test the model image with progress_bar=True.
    """
    pytest.importorskip('tqdm')
    params = QTable()
    params['x_0'] = [50, 70, 90]
    params['y_0'] = [50, 50, 50]
    params['gamma'] = [1.7, 2.32, 5.8]
    params['alpha'] = [2.9, 5.7, 4.6]
    model = Moffat2D(amplitude=1)
    shape = (300, 500)
    model_shape = (11, 11)
    image = make_model_image(shape, model, params, model_shape=model_shape,
                             progress_bar=True)
    assert image.shape == shape
    assert image.sum() > 1


def test_model_shape_from_bbox_no_bbox():
    """
    Test that _model_shape_from_bbox raises an error when the model does
    not have a bounding_box attribute.
    """
    model = Polynomial2D(degree=2)
    match = 'model does not have a bounding_box attribute'
    with pytest.raises(ValueError, match=match):
        _model_shape_from_bbox(model)
